import os

model_urls_root_r01 = (
    "https://github.com/jinfagang/wanwu_release/releases/download/v0.1"
)

LOCAL_CACHE_FOLDER = os.path.expanduser("~/wanwu_models")

"""
Following todo:

- SCRFD gpu model, fix efficientNMS issue;
- YOLOX CPU and GPU version model;

"""

model_urls = {
    # r10 version: init workable version with models
    "yolov5s_coco_trt": f"{model_urls_root_r01}/yolov5s_gpu.onnx",
    "yolov5s_coco_cpu": f"{model_urls_root_r01}/yolov5s_cpu.onnx",
    # "yolov5_s6_coco": f"{model_urls_root_r40}/yolov5_darknet_pan_s_r31_coco-eb728698.pt",
    "scrfd_500m_cpu": f"{model_urls_root_r01}/scrfd_500m_cpunms.onnx",
    "scrfd_500m_trt": f"{model_urls_root_r01}/scrfd_500m_gpu.onnx",
    "yolox_s_coco_trt": f"{model_urls_root_r01}/yolox_s_gpu.onnx",
    "yolox_tiny_coco_cpu": f"{model_urls_root_r01}/yolox_tiny_cpunms.onnx",
    "reid_osnet_x0_25_cpu": f"{model_urls_root_r01}/reid_osnet_x0_25_256x128.onnx",
    "reid_osnet_x0_25_trt": f"{model_urls_root_r01}/reid_osnet_x0_25_256x128.onnx",
    "alp_halpe26_res50_cpu": f"{model_urls_root_r01}/alp_halpe26_res50.onnx",
    "alp_halpe26_res50_trt": f"{model_urls_root_r01}/alp_halpe26_res50.onnx",
    "alp_halpe26_res18_cpu": f"{model_urls_root_r01}/alp_halpe26_r18.onnx",
    "alp_halpe26_res18_trt": f"{model_urls_root_r01}/alp_halpe26_r18.onnx",
    "yolov7_cpu": f"{model_urls_root_r01}/yolov7_cpu.onnx",
    "yolov7_trt": f"{model_urls_root_r01}/yolov7_gpu.onnx",
    "yolov7-tiny_cpu": f"{model_urls_root_r01}/yolov7-tiny_cpu.onnx",
    "yolov7-tiny_trt": f"{model_urls_root_r01}/yolov7-tiny_gpu.onnx",
    "yolov5n-0.5_cpu": f"{model_urls_root_r01}/yolov5n-0.5.onnx",
    "yolov7s-face_cpu": f"{model_urls_root_r01}/yolov7s-face.onnx",
}



from wanwu.utils.image_utils import load_inputs_std_mean
from ..core.infer import Backends, BaseInfer
from ..core.modelloader import ModelLoader, TaskType
import time
import numpy as np


class Classify(BaseInfer):

    def __init__(self, onnx_f: str = None, backend: Backends = Backends.CPU_ORT, type: str = 'yolov5s_coco',
                 batch_size=1, input_width=640, input_height=512,
                 num_classes=80, fp16=False, timing=False, **kargs) -> None:
        super(Classify, self).__init__(type=type, onnx_f=onnx_f, backend=backend,
                                       batch_size=batch_size, input_width=input_width, input_height=input_height)

        self.num_classes = num_classes

        self.timing = timing
        self.frame_time_cost = 0

        model_loader = ModelLoader(
            self.onnx_f, backend=self.backend, task_type=TaskType.CLS, fp16=fp16, num_classes=self.num_classes,
            batch_size=self.batch_size, input_width=self.input_width, input_height=self.input_height)
        self.infer_wrapper = model_loader.get_infer_wrapper()

    def infer(self, inp):
        if isinstance(inp, list):
            inp = np.concatenate(inp, axis=0)
        assert inp.shape[-2] == self.input_width or inp.shape[-2] == self.input_height, \
            f'Seems your inp shape not equal to your Det input hw, please check. {inp.shape} vs {self.input_height}x{self.input_width}'
        if self.timing:
            t0 = time.perf_counter()
        res = self.infer_wrapper.infer(inp)
        if self.timing:
            t1 = time.perf_counter()
            self.frame_time_cost = round(t1 - t0, 3)
        return res

    def load_inputs(self, img, std=1, mean=0, transpose=True, is_rgb=False):
        return load_inputs_std_mean(img, self.input_height, self.input_width,
                                    std=std, mean=mean, transpose=transpose, is_rgb=is_rgb
                                    )
    
    def vis_res(self, img, res, class_names=None):
        print('Classes: ', res)



from typing import OrderedDict
from alfred.deploy.tensorrt.common import (
    do_inference_v2,
    allocate_buffers_v2_dynamic,
    allocate_buffers_v2,
    load_engine_from_local,
    check_engine,
)
import numpy as np
from alfred.utils.log import logger
import time
from .base import WrapperBase, TensorInfo


class TensorRTInferencer(WrapperBase):
    def __init__(self, engine_f, device_id=0, cuda_ctx=None, timing=False) -> None:
        self.engine_f = engine_f
        self.device_id = device_id
        self.cuda_ctx = cuda_ctx
        if self.cuda_ctx:
            self.cuda_ctx.push()
        self.timing = timing
        self._init_engine()

    def _init_engine(self):
        self.engine = load_engine_from_local(self.engine_f)
        sps = check_engine(self.engine)
        self.input_shapes = [sps[i]["shape"] for i in sps.keys() if sps[i]["is_input"]]
        self.output_shapes = [
            sps[i]["shape"] for i in sps.keys() if not sps[i]["is_input"]
        ]
        self.inputs_dict = OrderedDict(
            {
                k: TensorInfo(sps[k]["shape"], "int")
                for k in sps.keys()
                if sps[k]["is_input"]
            }
        )
        self.outputs_dict = OrderedDict(
            {
                k: TensorInfo(sps[k]["shape"], "int")
                for k in sps.keys()
                if not sps[k]["is_input"]
            }
        )
        self.ori_input_shape = self.engine.get_binding_shape(0)
        self.is_dynamic_batch = self.ori_input_shape[0] == -1
        self.is_dynamic_shape = self.ori_input_shape[-1] == -1
        if self.is_dynamic_batch:
            logger.info(
                f"engine is dynamic on batch. engine max_batchsize: {self.engine.max_batch_size}"
            )
        if self.is_dynamic_shape:
            logger.info("engine is dynamic on shape.")

        try:
            self.context = self.engine.create_execution_context()
            if self.is_dynamic_shape:
                (
                    self.inputs,
                    self.outputs,
                    self.bindings,
                    self.stream,
                ) = allocate_buffers_v2_dynamic(self.engine)
                self.context.set_optimization_profile_async(0, self.stream.handle)
            else:
                (
                    self.inputs,
                    self.outputs,
                    self.bindings,
                    self.stream,
                ) = allocate_buffers_v2(self.engine)
            logger.info(f"TRT engine loaded. {self.input_shapes}")
            if self.cuda_ctx:
                self.cuda_ctx.pop()
        except Exception as e:
            self.cuda_ctx.pop()
            del self.cuda_ctx
            raise RuntimeError("Fail to allocate CUDA resources") from e

    def infer(self, imgs, return_dict=False):
        assert isinstance(imgs, np.ndarray), "imgs must be numpy array"
        if self.cuda_ctx:
            self.cuda_ctx.push()

        if self.is_dynamic_batch:
            bs = imgs.shape[0]
            self.inputs[0].host = imgs.ravel()
            # make context aware what's batchsize current
            self.ori_input_shape[0] = bs
            self.context.set_binding_shape(0, (self.ori_input_shape))
        else:
            self.inputs[0].host = imgs.ravel()

        if self.timing:
            t0 = time.perf_counter()
        outs = do_inference_v2(
            self.context,
            bindings=self.bindings,
            inputs=self.inputs,
            outputs=self.outputs,
            stream=self.stream,
            input_tensor=imgs,
        )
        if self.cuda_ctx:
            self.cuda_ctx.pop()
        if self.timing:
            t1 = time.perf_counter()
            print(f"engine cost: {t1 - t0}, fps: {1/(t1-t0)}")
        if not return_dict:
            outs_reshaped = []
            for i, o in enumerate(outs):
                o_s = self.output_shapes[i]
                if self.is_dynamic_batch:
                    o_s[0] = self.engine.max_batch_size
                o = np.reshape(o, o_s)
                if self.is_dynamic_batch:
                    outs_reshaped.append(o[:bs, ...])
                else:
                    outs_reshaped.append(o)
            return outs_reshaped
        else:
            outs_reshaped = {}
            for i, o in enumerate(outs):
                o_s = self.output_shapes[i]
                out_name = list(self.outputs_dict.keys())[i]
                if self.is_dynamic_batch:
                    o_s[0] = self.engine.max_batch_size
                o = np.reshape(o, o_s)
                if self.is_dynamic_batch:
                    outs_reshaped[out_name] = o[:bs, ...]
                else:
                    outs_reshaped[out_name] = o
            return outs_reshaped

    def __del__(self):
        """Free CUDA memories"""
        try:
            del self.outputs
            del self.inputs
            del self.stream
        except Exception as e:
            print(e)

    def get_inputs_dict(self):
        return self.inputs_dict

    def get_outputs_dict(self):
        return self.outputs_dict



# Copyright (c) 2022 Lucas Jin. All rights reserved.
#
# This source code is licensed under the GPL-3.0 license found in the
# LICENSE file in the root directory of this source tree.
#
# Copyright (c) 2021, NVIDIA CORPORATION. All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in the
# LICENSE file in the root directory of TensorRT source tree.
#

from pathlib import Path

import numpy as np
import onnx
from onnx import shape_inference
from typing import Any, List, Text, Dict, Set
from onnx import ModelProto, ValueInfoProto

import onnx.checker
from alfred.utils.log import logger

try:
    import onnx_graphsurgeon as gs
except ImportError:
    logger.warning('onnx_graphsurgeon is not installed, not do surgeon.')
    gs = None

"""
this file try surgeon on normal Detection onnx model
and add BatchedNMS plugin to it.
"""


# type: (ModelProto, Dict[Text, List[Any]], Dict[Text, List[Any]]) -> ModelProto
def update_inputs_outputs_dims(model, input_dims, output_dims=None):
    """
        This function updates the dimension sizes of the model's inputs and outputs to the values
        provided in input_dims and output_dims. if the dim value provided is negative, a unique dim_param
        will be set for that dimension.

        Example. if we have the following shape for inputs and outputs:
                shape(input_1) = ('b', 3, 'w', 'h')
                shape(input_2) = ('b', 4)
                and shape(output)  = ('b', 'd', 5)

            The parameters can be provided as:
                input_dims = {
                    "input_1": ['b', 3, 'w', 'h'],
                    "input_2": ['b', 4],
                }
                output_dims = {
                    "output": ['b', -1, 5]
                }

            Putting it together:
                model = onnx.load('model.onnx')
                updated_model = update_inputs_outputs_dims(model, input_dims, output_dims)
                onnx.save(updated_model, 'model.onnx')
    """
    dim_param_set = set()  # type: Set[Text]

    # type: (Set[Text], List[ValueInfoProto]) -> None
    def init_dim_param_set(dim_param_set, value_infos):
        for info in value_infos:
            shape = info.type.tensor_type.shape
            for dim in shape.dim:
                if dim.HasField('dim_param'):
                    dim_param_set.add(dim.dim_param)  # type: ignore

    init_dim_param_set(dim_param_set, model.graph.input)  # type: ignore
    init_dim_param_set(dim_param_set, model.graph.output)  # type: ignore
    init_dim_param_set(dim_param_set, model.graph.value_info)  # type: ignore

    def update_dim(tensor, dim, j, name):  # type: (ValueInfoProto, Any, int, Text) -> None
        dim_proto = tensor.type.tensor_type.shape.dim[j]
        if isinstance(dim, int):
            if dim >= 0:
                if dim_proto.HasField('dim_value') and dim_proto.dim_value != dim:
                    raise ValueError('Unable to set dimension value to {} for axis {} of {}. Contradicts existing dimension value {}.'
                                     .format(dim, j, name, dim_proto.dim_value))
                dim_proto.dim_value = dim
            else:
                generated_dim_param = name + '_' + str(j)
                if generated_dim_param in dim_param_set:
                    raise ValueError('Unable to generate unique dim_param for axis {} of {}. Please manually provide a dim_param value.'
                                     .format(j, name))
                dim_proto.dim_param = generated_dim_param
        elif isinstance(dim, str):
            dim_proto.dim_param = dim
        else:
            raise ValueError(
                'Only int or str is accepted as dimension value, incorrect type: {}'.format(type(dim)))
    
    new_batch_size = 1
    for input in model.graph.input:
        input_name = input.name
        input_dim_arr = input_dims[input_name]
        new_batch_size = input_dim_arr[0]
        for j, dim in enumerate(input_dim_arr):
            update_dim(input, dim, j, input_name)

    if output_dims is not None:
        for output in model.graph.output:
            output_name = output.name
            output_dim_arr = output_dims[output_name]
            for j, dim in enumerate(output_dim_arr):
                update_dim(output, dim, j, output_name)
    else:
        # only update output on batchSize
        logger.info('updating onnx model output batchsize to: {}'.format(new_batch_size))
        for output in model.graph.output:
            output_name = output.name
            update_dim(output, new_batch_size, 0, output_name)
    onnx.checker.check_model(model)
    return model


def get_inputs(model: onnx.ModelProto) -> List[onnx.ValueInfoProto]:
    initializer_names = [x.name for x in model.graph.initializer]
    return [ipt for ipt in model.graph.input if ipt.name not in initializer_names]


def is_model_dynamic_on_hw(model):
    _inputs = get_inputs(model)
    input_specs = dict()

    for _input in _inputs:
        tensor_type = _input.type.tensor_type
        sh = []
        # check if it has a shape:
        if (tensor_type.HasField("shape")):
            # iterate through dimensions of the shape:
            for d in tensor_type.shape.dim:
                # the dimension may have a definite (integer) value or a symbolic identifier or neither:
                if (d.HasField("dim_value")):
                    sh.append(d.dim_value)
                elif (d.HasField("dim_param")):
                    sh.append(-1)
                else:
                    sh.append('?')
            input_specs[_input.name] = sh
        else:
            print("unknown rank", end="")
    # logger.info(f'input shapes: {input_specs}')
    # check input0
    a = list(input_specs.values())[0]
    if a[-2] == -1:
        return True
    else:
        return False


def is_model_dynamic_on_batch(onnx_model):
    _inputs = get_inputs(onnx_model)
    input_specs = dict()

    for _input in _inputs:
        tensor_type = _input.type.tensor_type
        sh = []
        # check if it has a shape:
        if (tensor_type.HasField("shape")):
            # iterate through dimensions of the shape:
            for d in tensor_type.shape.dim:
                # the dimension may have a definite (integer) value or a symbolic identifier or neither:
                if (d.HasField("dim_value")):
                    sh.append(d.dim_value)
                elif (d.HasField("dim_param")):
                    sh.append(-1)
                else:
                    sh.append('?')
            input_specs[_input.name] = sh
        else:
            print("unknown rank", end="")
    a = list(input_specs.values())[0]
    if a[0] == -1:
        return True, input_specs
    else:
        return False, None


class OnnxModelSurgeonOperationer:
    """
    Surgeon normal detection model, attach different NMS plugin based on backend
    """

    def __init__(
        self,
        onnx_path: str,
        num_classes: int,
        enable_dynamic: bool = True,
        batch_size: int = 1,
    ):
        onnx_path = Path(onnx_path)
        assert onnx_path.exists()

        self.num_classes = num_classes

        self.onnx_path = onnx_path
        logger.info(f"Loaded origin onnx model from {onnx_path}")

        self.onnx_model = onnx.load(onnx_path)
        self.is_dynamic_on_hw = is_model_dynamic_on_hw(self.onnx_model)
        self.is_dynamic_on_batch, self.inputs_dict = is_model_dynamic_on_batch(self.onnx_model)
        
        if gs is not None:
            self.graph = gs.import_onnx(self.onnx_model)
            assert self.graph
            logger.info("ONNX graph created successfully")

            # Fold constants via ONNX-GS that PyTorch2ONNX may have missed
            self.graph.fold_constants()
            self.batch_size = batch_size

            self.input_name = self.graph.inputs[0].name
            self.ok = True
        else:
            self.ok = False

    def infer(self):
        """
        Sanitize the graph by cleaning any unconnected nodes, do a topological resort,
        and fold constant inputs values. When possible, run shape inference on the
        ONNX graph to determine tensor shapes.
        """
        for _ in range(3):
            count_before = len(self.graph.nodes)

            self.graph.cleanup().toposort()
            try:
                for node in self.graph.nodes:
                    for o in node.outputs:
                        o.shape = None
                model = gs.export_onnx(self.graph)
                model = shape_inference.infer_shapes(model)
                self.graph = gs.import_onnx(model)
            except Exception as e:
                logger.info(
                    f"Shape inference could not be performed at this time:\n{e}")
            try:
                self.graph.fold_constants(fold_shapes=True)
            except TypeError as e:
                logger.error(
                    "This version of ONNX GraphSurgeon does not support folding shapes, "
                    f"please upgrade your onnx_graphsurgeon module. Error:\n{e}"
                )
                raise

            count_after = len(self.graph.nodes)
            if count_before == count_after:
                # No new folding occurred in this iteration, so we can stop for now.
                break

    def save(self, output_path=None):
        """
        Save the ONNX model to the given location.

        Args:
            output_path: Path pointing to the location where to write
                out the updated ONNX model.
        """
        if output_path is None:
            output_path = self.onnx_path.with_name(
                self.onnx_path.stem + '_w_nms' + self.onnx_path.suffix)
        if gs is not None:
            self.graph.cleanup().toposort()
            model = gs.export_onnx(self.graph)
            onnx.save(model, output_path)
            logger.info(f"Saved ONNX model to {output_path}")
            return output_path, model
        else:
            return output_path, self.onnx_model

    def update_inputs_dims(self, batch_size=1, input_height=512, input_width=608):
        if self.is_dynamic_on_hw:
            updated_model = update_inputs_outputs_dims(self.onnx_model,
                                                       {self.input_name: [batch_size, 3, input_height, input_width]})
            # update surgeon graph
            self.graph = gs.import_onnx(updated_model)
            logger.info('Updated model inputs...')
        # else:
        #     logger.info(
        #         '.')

    def register_trt_batched_nms(
        self,
        score_thresh: float = 0.25,
        nms_thresh: float = 0.45,
        detections_per_img: int = 100,
        normalized: bool = True,
    ):
        """
        Register the ``BatchedNMS_TRT`` plugin node.

        NMS expects these shapes for its input tensors:
            - box_net: [batch_size, number_boxes, 1, 4]
            - class_net: [batch_size, number_boxes, number_labels]

        Args:
            score_thresh (float): The scalar threshold for score (low scoring boxes are removed).
            nms_thresh (float): The scalar threshold for IOU (new boxes that have high IOU
                overlap with previously selected boxes are removed).
            detections_per_img (int): Number of best detections to keep after NMS.
            normalized (bool): Set to false if the box coordinates are not normalized,
                meaning they are not in the range [0,1]. Defaults: True.
        """

        self.infer()
        # Find the concat node at the end of the network
        nms_inputs = self.graph.outputs
        op = "BatchedNMS_TRT"
        attrs = {
            "plugin_version": "1",
            "shareLocation": True,
            "backgroundLabelId": -1,  # no background class
            "numClasses": self.num_classes,
            "topK": 1024,
            "keepTopK": detections_per_img,
            "scoreThreshold": score_thresh,
            "iouThreshold": nms_thresh,
            "isNormalized": normalized,
            "clipBoxes": False,
        }

        # NMS Outputs
        output_num_detections = gs.Variable(
            name="num_detections",
            dtype=np.int32,
            shape=[self.batch_size, 1],
        )  # A scalar indicating the number of valid detections per batch image.
        output_boxes = gs.Variable(
            name="detection_boxes",
            dtype=np.float32,
            shape=[self.batch_size, detections_per_img, 4],
        )
        output_scores = gs.Variable(
            name="detection_scores",
            dtype=np.float32,
            shape=[self.batch_size, detections_per_img],
        )
        output_labels = gs.Variable(
            name="detection_classes",
            dtype=np.float32,
            shape=[self.batch_size, detections_per_img],
        )

        nms_outputs = [output_num_detections,
                       output_boxes, output_scores, output_labels]

        # Create the NMS Plugin node with the selected inputs. The outputs of the node will also
        # become the final outputs of the graph.
        self.graph.layer(
            op=op,
            name="batched_nms",
            inputs=nms_inputs,
            outputs=nms_outputs,
            attrs=attrs,
        )
        logger.info(
            f"Created NMS plugin '{op}' with attributes: {attrs}, using BatchSize: {self.batch_size}")
        self.graph.outputs = nms_outputs
        self.infer()

    def register_trt_efficient_nms(
        self,
        score_thresh: float = 0.25,
        nms_thresh: float = 0.45,
        detections_per_img: int = 100,
    ):
        """
        Register the ``EfficientNMS_TRT`` plugin node.

        NMS expects these shapes for its input tensors:
            - box_net: [batch_size, number_boxes, 4]
            - class_net: [batch_size, number_boxes, number_labels]

        Args:
            score_thresh (float): The scalar threshold for score (low scoring boxes are removed).
            nms_thresh (float): The scalar threshold for IOU (new boxes that have high IOU
                overlap with previously selected boxes are removed).
            detections_per_img (int): Number of best detections to keep after NMS.
        """

        self.infer()
        # Find the concat node at the end of the network
        nms_inputs = self.graph.outputs

        op = "EfficientNMS_TRT"
        attrs = {
            "plugin_version": "1",
            "background_class": -1,  # no background class
            "max_output_boxes": detections_per_img,
            "score_threshold": score_thresh,
            "iou_threshold": nms_thresh,
            "score_activation": False,
            "box_coding": 0,
        }

        # NMS Outputs
        output_num_detections = gs.Variable(
            name="num_detections",
            dtype=np.int32,
            shape=[self.batch_size, 1],
        )  # A scalar indicating the number of valid detections per batch image.
        output_boxes = gs.Variable(
            name="detection_boxes",
            dtype=np.float32,
            shape=[self.batch_size, detections_per_img, 4],
        )
        output_scores = gs.Variable(
            name="detection_scores",
            dtype=np.float32,
            # dtype=np.float16,
            shape=[self.batch_size, detections_per_img],
        )
        output_labels = gs.Variable(
            name="detection_classes",
            dtype=np.int32,
            shape=[self.batch_size, detections_per_img],
        )

        nms_outputs = [output_num_detections,
                       output_boxes, output_scores, output_labels]

        # Create the NMS Plugin node with the selected inputs. The outputs of the node will also
        # become the final outputs of the graph.
        self.graph.layer(
            op=op,
            name="batched_nms",
            inputs=nms_inputs,
            outputs=nms_outputs,
            attrs=attrs,
        )
        logger.info(
            f"Created Efficient NMS plugin '{op}' with attributes: {attrs}")
        self.graph.outputs = nms_outputs
        self.infer()

    def register_onnx_nonmaxsuppression(
        self,
        score_thresh: float = 0.25,
        iou_threshold: float = 0.45,
        max_output_boxes_per_class: int = 100,
        normalized: bool = True,
    ):
        """
        Register the ``NonMaxSuppression`` node

        WARN: this is not supported yet, seems it using different inputs with TRT

        NMS expects these shapes for its input tensors:
            - boxes: [batch_size, number_boxes, 4]
            - scores: [batch_size, number_labels, number_boxes]
        the scores last 2 dimension is opposite with EfficientNMS
        """

        self.infer()
        # Find the concat node at the end of the network
        nms_inputs = self.graph.outputs
        op = "NonMaxSuppression"
        attrs = {
            "plugin_version": "1",
            "shareLocation": True,
            "backgroundLabelId": -1,  # no background class
            "numClasses": self.num_classes,
            "topK": 1024,
            "max_output_boxes_per_class": max_output_boxes_per_class,
            "score_thresh": score_thresh,
            "iou_threshold": iou_threshold,
            "isNormalized": normalized,
            "clipBoxes": False,
        }

        # NMS Outputs
        output_num_detections = gs.Variable(
            name="num_detections",
            dtype=np.int32,
            shape=[self.batch_size, 1],
        )  # A scalar indicating the number of valid detections per batch image.
        output_boxes = gs.Variable(
            name="detection_boxes",
            dtype=np.float32,
            shape=[self.batch_size, max_output_boxes_per_class, 4],
        )
        output_scores = gs.Variable(
            name="detection_scores",
            dtype=np.float32,
            shape=[self.batch_size, max_output_boxes_per_class],
        )
        output_labels = gs.Variable(
            name="detection_classes",
            dtype=np.float32,
            shape=[self.batch_size, max_output_boxes_per_class],
        )

        nms_outputs = [output_num_detections,
                       output_boxes, output_scores, output_labels]

        # Create the NMS Plugin node with the selected inputs. The outputs of the node will also
        # become the final outputs of the graph.
        self.graph.layer(
            op=op,
            name="batched_nms",
            inputs=nms_inputs,
            outputs=nms_outputs,
            attrs=attrs,
        )
        logger.info(f"Created NMS plugin '{op}' with attributes: {attrs}")

        self.graph.outputs = nms_outputs

        self.infer()



import cv2
from facemesh import FaceMesh
from blazeface import BlazeFace
import torch


# load FaceMesh model
mesh_net = FaceMesh()
mesh_net.load_weights("facemesh.pth")
# load BlazeFace model
blaze_net = BlazeFace()
blaze_net.load_weights("blazeface.pth")
blaze_net.load_anchors("anchors.npy")
# postprocessing for face detector

torch.onnx.export(blaze_net, torch.randn(1, 3, 128, 128, device='cpu'), "blazeface.onnx",
                  input_names=("image", ), output_names=("preds", "confs"), opset_version=9
                  )

torch.onnx.export(mesh_net, torch.randn(1, 3, 192, 192, device='cpu'), "facemesh.onnx",
                  input_names=("image", ), output_names=("preds", "confs"), opset_version=9
                  )


def get_crop_face(detections, image):
    w, h = image.shape[0], image.shape[1]
    ymin = int(detections[0, 0] * w)
    xmin = int(detections[0, 1] * h)
    ymax = int(detections[0, 2] * w)
    xmax = int(detections[0, 3] * h)
    margin_x = int(0.25 * (xmax - xmin))
    margin_y = int(0.25 * (ymax - ymin))
    ymin -= margin_y
    ymax += margin_y
    xmin -= margin_x
    xmax += margin_x
    face_img = image[ymin:ymax, xmin:xmax]
    face_img = cv2.cvtColor(face_img, cv2.COLOR_BGR2RGB)
    image = cv2.rectangle(image, (xmin, ymin), (xmax, ymax), (0, 0, 255), 2)
    return xmin, ymin, face_img
# postprocessing for mesh


def get_mesh_face(detections, face_img, image, xmin, ymin):
    xscale, yscale = 192 / face_img.shape[1], 192 / face_img.shape[0]
    for i in range(detections.shape[0]):
        x, y = int(detections[i, 0] / xscale), int(detections[i, 1] / yscale)
        image = cv2.circle(image, (xmin + x, ymin + y), 1, (255, 0, 0), 1)


videoCapture = cv2.VideoCapture(0)
while True:
    ret, image = videoCapture.read()
    if not ret:
        break
    # preprocess image
    face_img = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    face_img = cv2.resize(face_img, (128, 128))
    # get face detection boxes
    detections = blaze_net.predict_on_image(face_img).numpy()
    xmin, ymin, face_img = get_crop_face(detections, image)
    # get face mesh
    mesh_img = cv2.resize(face_img, (192, 192))
    detections = mesh_net.predict_on_image(mesh_img).numpy()
    get_mesh_face(detections, face_img, image, xmin, ymin)
    # show processed image
    cv2.imshow("capture", image)
    if cv2.waitKey(3) & 0xFF == 27:
        break



'''
auto detect backend to use
logic:

if GPU:
    if tensorrt installed:
        tensorrt
    else:
        ort gpu
else:
    if openvino installed:
        openvino
    else:
        ort
'''
from ast import Import
from math import fabs
from typing import List
import wanwu.utils.gpu_utils as gpu_utils
from wanwu.core.enums import Backends
from alfred.utils.log import logger


def print_gpus_info(gpus: List[gpu_utils.GPU]):
    logger.info('Located GPU info: ')
    for g in gpus:
        logger.info(f'  {g.name}, {g.memoryUsed}MB/{g.memoryTotal}MB')


def dispatcher_on_device_auto():
    """
    return the fastest backend based on currently hardware and software
    """
    gpus = gpu_utils.getGPUs()
    if len(gpus) > 0:
        # check if tensorrt installed
        print_gpus_info(gpus)
        try:
            import tensorrt
            logger.info('Auto switching backend to TensorRT.')
            return Backends.GPU_TENSORRT
        except ImportError:
            # tensorrt not install
            logger.info(
                'Not found TensorRT installed, fallback to default CUDA, this might not fastest.')
            return Backends.GPU_ORT
    else:
        try:
            import openvino
            return Backends.CPU_OPENVINO
        except ImportError:
            logger.info('openvino not installed, fallback to normal CPU.')
            return Backends.CPU_ORT


def check_backend_support_on_host(b: Backends):
    if b == Backends.GPU_TENSORRT:
        try:
            from wanwu.core.backends.trt import TensorRTInferencer
            return True
        except ImportError as e:
            logger.warning(f'import trt error: {e}')
            return False
    elif b == Backends.CPU_ORT:
        try:
            import onnxruntime
            return True
        except ImportError:
            return False
    elif b == Backends.CPU_MNN:
        try:
            import MNN
            return True
        except ImportError:
            return False
    elif b == Backends.CPU_PPLNN:
        try:
            import pyppl
            return True 
        except ImportError:
            return False
    elif b == Backends.CPU_OPENVINO:
        try:
            from openvino import runtime
            return True
        except ImportError:
            return False
    return False



import numpy as np
from alfred.utils.log import logger
from wanwu.core.backends.base import TensorInfo
import os

try:
    import onnxruntime as rt
    import onnxruntime
except ImportError:
    onnxruntime = None
from alfred.utils.log import logger
from .base import WrapperBase


def ort_dtype_to_numpy(od):
    if "float" in od: 
        return np.float32
    elif "int" in od:
        return np.int64


class ORTWrapper(WrapperBase):
    def __init__(self, onnx_f) -> None:
        self.onnx_f = onnx_f
        so = onnxruntime.SessionOptions()
        so.intra_op_num_threads = 5
        so.execution_mode = onnxruntime.ExecutionMode.ORT_PARALLEL
        so.inter_op_num_threads = 5
        so.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_ENABLE_ALL
        os.environ["OMP_NUM_THREADS"] = "5"
        self.ort_session = onnxruntime.InferenceSession(onnx_f, sess_options=so)
        self.input_names = [input.name for input in self.ort_session.get_inputs()]
        # todo: extract possible input hw for vision models

    def infer(self, inputs, return_dict=True):
        if isinstance(inputs, dict):
            ort_inputs = inputs
        else:
            if not isinstance(inputs, list):
                # incase users may input single img
                inputs = [inputs]
            assert len(inputs) == len(
                self.ort_session.get_inputs()
            ), "inputs must same with model."
            ort_inputs = dict(
                (self.ort_session.get_inputs()[i].name, inpt)
                for i, inpt in enumerate(inputs)
            )
        ort_outs = self.ort_session.run(None, ort_inputs)

        if return_dict:
            outs_dict = dict()
            for i, oo in enumerate(self.ort_session.get_outputs()):
                n = oo.name
                outs_dict[n] = ort_outs[i]
            return outs_dict
        else:
            return ort_outs

    def get_inputs_dict(self):
        input_tensor_dict = {}
        inputs_node = self.ort_session.get_inputs()
        for inode in inputs_node:
            sh = inode.shape
            # handle dynamic shapes
            sh_replaced = []
            for i in sh:
                if isinstance(i, str):
                    sh_replaced.append(-1)
                else:
                    sh_replaced.append(i)
            input_tensor_dict[inode.name] = TensorInfo(
                sh_replaced, ort_dtype_to_numpy(inode.type)
            )
        return input_tensor_dict

    def get_outputs_dict(self):
        output_tensor_dict = {}
        inputs_node = self.ort_session.get_outputs()
        for inode in inputs_node:
            sh = inode.shape
            sh_replaced = []
            for i in sh:
                if isinstance(i, str):
                    sh_replaced.append(-1)
                else:
                    sh_replaced.append(i)
            output_tensor_dict[inode.name] = TensorInfo(sh_replaced, "float")
        return output_tensor_dict





from wanwu.utils.image_utils import load_inputs_std_mean
from ..core.infer import Backends, BaseInfer
from ..core.modelloader import ModelLoader, TaskType
import time
import numpy as np


class Seg(BaseInfer):
    def __init__(
        self,
        onnx_f: str = None,
        backend: Backends = Backends.CPU_ORT,
        type: str = "yolov5s_coco",
        batch_size=1,
        input_width=640,
        input_height=512,
        num_classes=80,
        fp16=False,
        timing=False,
        **kargs,
    ) -> None:
        super(Seg, self).__init__(
            type=type,
            onnx_f=onnx_f,
            backend=backend,
            batch_size=batch_size,
            input_width=input_width,
            input_height=input_height,
        )

        # add some CPU realtime Segmentation Models
        self.num_classes = num_classes

        self.timing = timing
        self.frame_time_cost = 0

        self.model_loader = ModelLoader(
            self.onnx_f,
            backend=self.backend,
            task_type=TaskType.SEG,
            fp16=fp16,
            num_classes=self.num_classes,
            batch_size=self.batch_size,
            input_width=self.input_width,
            input_height=self.input_height,
        )
        self.infer_wrapper = self.model_loader.get_infer_wrapper()

    def infer(self, inp):
        if not self.model_loader.is_dynamic_on_hw:
            assert (
                inp.shape[-2] == self.input_width or inp.shape[-2] == self.input_height
            ), f"Seems your inp shape not equal to your Det input hw, please check. {inp.shape} vs {self.input_height}x{self.input_width}"
        if self.timing:
            t0 = time.perf_counter()
        res = self.infer_wrapper.infer(inp)
        if self.timing:
            t1 = time.perf_counter()
            self.frame_time_cost = round(t1 - t0, 3)
        return res

    def load_inputs(self, img, std, mean, transpose=True, is_rgb=False):
        return load_inputs_std_mean(
            img,
            self.input_height,
            self.input_width,
            std=std,
            mean=mean,
            transpose=transpose,
            is_rgb=is_rgb,
        )

    def vis_res(self, img, res, class_names=None):
        print("Classes: ", res)



import numpy as np


def ltwh_to_x1y1x2y2(boxes):
    # Ensure the input is a numpy array
    boxes = np.array(boxes)

    # Create a new array for the output
    new_boxes = np.zeros_like(boxes)

    # Set x1 and y1
    new_boxes[:, 0] = boxes[:, 0]  # x1 = left
    new_boxes[:, 1] = boxes[:, 1]  # y1 = top

    # Set x2 and y2
    new_boxes[:, 2] = boxes[:, 0] + boxes[:, 2]  # x2 = left + width
    new_boxes[:, 3] = boxes[:, 1] + boxes[:, 3]  # y2 = top + height

    return new_boxes



import numpy as np
from alfred.utils.log import logger
from wanwu.core.backends.base import TensorInfo
import os

try:
    from openvino import runtime
except ImportError:
    runtime = None
from alfred.utils.log import logger
from .base import WrapperBase

try:
    import onnxruntime as rt
    import onnxruntime
except ImportError:
    onnxruntime = None


def ort_dtype_to_numpy(od):
    if "float" in od:
        return np.float32
    elif "int" in od:
        return np.int64


class OpenVINOWrapper(WrapperBase):
    def __init__(self, onnx_f) -> None:
        self.onnx_f = onnx_f
        os.environ["OMP_NUM_THREADS"] = "5"

        self.openvino_executor = runtime.Core()
        self.model = self.openvino_executor.compile_model(
            model=self.openvino_executor.read_model(model=self.onnx_f),
            device_name="CPU",
        )

        ort_session = onnxruntime.InferenceSession(onnx_f)

        self.inputs_dict = {}
        self.outputs_dict = {}

        # OpenVINO can not get dynamic shape
        # infer_request = self.model.create_infer_request()
        # for output in infer_request.model_outputs:
        #     sh = output.shape
        #     dt = output.dtype
        #     self.outputs_dict[output.any_name] = TensorInfo(sh, dt)

        # for inp in infer_request.model_inputs:
        #     sh = inp.shape
        #     dt = inp.dtype
        #     self.inputs_dict[inp.any_name] = TensorInfo(sh, dt)

        inputs_node = ort_session.get_inputs()
        for inode in inputs_node:
            sh = inode.shape
            # handle dynamic shapes
            sh_replaced = []
            for i in sh:
                if isinstance(i, str):
                    sh_replaced.append(-1)
                else:
                    sh_replaced.append(i)
            self.inputs_dict[inode.name] = TensorInfo(
                sh_replaced, ort_dtype_to_numpy(inode.type)
            )

        inputs_node = ort_session.get_outputs()
        for inode in inputs_node:
            sh = inode.shape
            sh_replaced = []
            for i in sh:
                if isinstance(i, str):
                    sh_replaced.append(-1)
                else:
                    sh_replaced.append(i)
            self.outputs_dict[inode.name] = TensorInfo(sh_replaced, "float")

        del ort_session

        logger.info(
            f"inputnames: {self.inputs_dict.keys()}, outputnames: {self.outputs_dict.keys()}"
        )
        # todo: extract possible input hw for vision models

    def infer(self, inputs, return_dict=False):
        if isinstance(inputs, dict):
            ort_inputs = inputs
        else:
            if not isinstance(inputs, list):
                # incase users may input single img
                inputs = [inputs]
            assert len(inputs) == len(
                self.ort_session.get_inputs()
            ), "inputs must same with model."
            ort_inputs = inputs

        self.infer_request = self.model.create_infer_request()
        ort_outs = self.infer_request.infer(ort_inputs)
        return ort_outs

    def get_inputs_dict(self):
        return self.inputs_dict

    def get_outputs_dict(self):
        return self.outputs_dict




# -*- coding: utf-8 -*-

'''
name:       demo_mnn.py
date:       2020-12-16 11:21:07
Env.:       Python 3.7.3, WIN 10
'''

import argparse
from pathlib import Path
from abc import ABCMeta, abstractmethod

import cv2
import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm
from scipy.special import softmax


# Copy from nanodet/util/visualization.py
_COLORS = np.array(
    [
        0.000, 0.447, 0.741,
        0.850, 0.325, 0.098,
        0.929, 0.694, 0.125,
        0.494, 0.184, 0.556,
        0.466, 0.674, 0.188,
        0.301, 0.745, 0.933,
        0.635, 0.078, 0.184,
        0.300, 0.300, 0.300,
        0.600, 0.600, 0.600,
        1.000, 0.000, 0.000,
        1.000, 0.500, 0.000,
        0.749, 0.749, 0.000,
        0.000, 1.000, 0.000,
        0.000, 0.000, 1.000,
        0.667, 0.000, 1.000,
        0.333, 0.333, 0.000,
        0.333, 0.667, 0.000,
        0.333, 1.000, 0.000,
        0.667, 0.333, 0.000,
        0.667, 0.667, 0.000,
        0.667, 1.000, 0.000,
        1.000, 0.333, 0.000,
        1.000, 0.667, 0.000,
        1.000, 1.000, 0.000,
        0.000, 0.333, 0.500,
        0.000, 0.667, 0.500,
        0.000, 1.000, 0.500,
        0.333, 0.000, 0.500,
        0.333, 0.333, 0.500,
        0.333, 0.667, 0.500,
        0.333, 1.000, 0.500,
        0.667, 0.000, 0.500,
        0.667, 0.333, 0.500,
        0.667, 0.667, 0.500,
        0.667, 1.000, 0.500,
        1.000, 0.000, 0.500,
        1.000, 0.333, 0.500,
        1.000, 0.667, 0.500,
        1.000, 1.000, 0.500,
        0.000, 0.333, 1.000,
        0.000, 0.667, 1.000,
        0.000, 1.000, 1.000,
        0.333, 0.000, 1.000,
        0.333, 0.333, 1.000,
        0.333, 0.667, 1.000,
        0.333, 1.000, 1.000,
        0.667, 0.000, 1.000,
        0.667, 0.333, 1.000,
        0.667, 0.667, 1.000,
        0.667, 1.000, 1.000,
        1.000, 0.000, 1.000,
        1.000, 0.333, 1.000,
        1.000, 0.667, 1.000,
        0.333, 0.000, 0.000,
        0.500, 0.000, 0.000,
        0.667, 0.000, 0.000,
        0.833, 0.000, 0.000,
        1.000, 0.000, 0.000,
        0.000, 0.167, 0.000,
        0.000, 0.333, 0.000,
        0.000, 0.500, 0.000,
        0.000, 0.667, 0.000,
        0.000, 0.833, 0.000,
        0.000, 1.000, 0.000,
        0.000, 0.000, 0.167,
        0.000, 0.000, 0.333,
        0.000, 0.000, 0.500,
        0.000, 0.000, 0.667,
        0.000, 0.000, 0.833,
        0.000, 0.000, 1.000,
        0.000, 0.000, 0.000,
        0.143, 0.143, 0.143,
        0.286, 0.286, 0.286,
        0.429, 0.429, 0.429,
        0.571, 0.571, 0.571,
        0.714, 0.714, 0.714,
        0.857, 0.857, 0.857,
        0.000, 0.447, 0.741,
        0.314, 0.717, 0.741,
        0.50, 0.5, 0
    ]
).astype(np.float32).reshape(-1, 3)


def get_resize_matrix(raw_shape, dst_shape, keep_ratio):
    """
    Get resize matrix for resizing raw img to input size
    :param raw_shape: (width, height) of raw image
    :param dst_shape: (width, height) of input image
    :param keep_ratio: whether keep original ratio
    :return: 3x3 Matrix
    """
    r_w, r_h = raw_shape
    d_w, d_h = dst_shape
    Rs = np.eye(3)
    if keep_ratio:
        C = np.eye(3)
        C[0, 2] = - r_w / 2
        C[1, 2] = - r_h / 2

        if r_w / r_h < d_w / d_h:
            ratio = d_h / r_h
        else:
            ratio = d_w / r_w
        Rs[0, 0] *= ratio
        Rs[1, 1] *= ratio

        T = np.eye(3)
        T[0, 2] = 0.5 * d_w
        T[1, 2] = 0.5 * d_h
        return T @ Rs @ C
    else:
        Rs[0, 0] *= d_w / r_w
        Rs[1, 1] *= d_h / r_h
        return Rs


def warp_boxes(boxes, M, width, height):
    """Apply transform to boxes
    Copy from nanodet/data/transform/warp.py
    """
    n = len(boxes)
    if n:
        # warp points
        xy = np.ones((n * 4, 3))
        xy[:, :2] = boxes[:, [0, 1, 2, 3, 0, 3, 2, 1]].reshape(
            n * 4, 2)  # x1y1, x2y2, x1y2, x2y1
        xy = xy @ M.T  # transform
        xy = (xy[:, :2] / xy[:, 2:3]).reshape(n, 8)  # rescale
        # create new boxes
        x = xy[:, [0, 2, 4, 6]]
        y = xy[:, [1, 3, 5, 7]]
        xy = np.concatenate(
            (x.min(1), y.min(1), x.max(1), y.max(1))).reshape(4, n).T
        # clip boxes
        xy[:, [0, 2]] = xy[:, [0, 2]].clip(0, width)
        xy[:, [1, 3]] = xy[:, [1, 3]].clip(0, height)
        return xy.astype(np.float32)
    else:
        return boxes


def overlay_bbox_cv(img, all_box, class_names):
    """Draw result boxes
    Copy from nanodet/util/visualization.py
    """
    # all_box array of [label, x0, y0, x1, y1, score]
    all_box.sort(key=lambda v: v[5])
    for box in all_box:
        label, x0, y0, x1, y1, score = box
        # color = self.cmap(i)[:3]
        color = (_COLORS[label] * 255).astype(np.uint8).tolist()
        text = '{}:{:.1f}%'.format(class_names[label], score * 100)
        txt_color=(0, 0, 0) if np.mean(_COLORS[label]) > 0.5 else (255, 255, 255)
        font = cv2.FONT_HERSHEY_SIMPLEX
        txt_size = cv2.getTextSize(text, font, 0.5, 2)[0]
        cv2.rectangle(img, (x0, y0), (x1, y1), color, 2)

        cv2.rectangle(img,
                      (x0, y0 - txt_size[1] - 1),
                      (x0 + txt_size[0] + txt_size[1], y0 - 1), color, -1)
        cv2.putText(img, text, (x0, y0-1),
                    font, 0.5, txt_color, thickness=1)
    return img


def hard_nms(box_scores, iou_threshold, top_k=-1, candidate_size=200):
    """

    Args:
        box_scores (N, 5): boxes in corner-form and probabilities.
        iou_threshold: intersection over union threshold.
        top_k: keep top_k results. If k <= 0, keep all the results.
        candidate_size: only consider the candidates with the highest scores.
    Returns:
         picked: a list of indexes of the kept boxes
    """
    scores = box_scores[:, -1]
    boxes = box_scores[:, :-1]
    picked = []
    # _, indexes = scores.sort(descending=True)
    indexes = np.argsort(scores)
    # indexes = indexes[:candidate_size]
    indexes = indexes[-candidate_size:]
    while len(indexes) > 0:
        # current = indexes[0]
        current = indexes[-1]
        picked.append(current)
        if 0 < top_k == len(picked) or len(indexes) == 1:
            break
        current_box = boxes[current, :]
        # indexes = indexes[1:]
        indexes = indexes[:-1]
        rest_boxes = boxes[indexes, :]
        iou = iou_of(
            rest_boxes,
            np.expand_dims(current_box, axis=0),
        )
        indexes = indexes[iou <= iou_threshold]

    return box_scores[picked, :]


def iou_of(boxes0, boxes1, eps=1e-5):
    """Return intersection-over-union (Jaccard index) of boxes.

    Args:
        boxes0 (N, 4): ground truth boxes.
        boxes1 (N or 1, 4): predicted boxes.
        eps: a small number to avoid 0 as denominator.
    Returns:
        iou (N): IoU values.
    """
    overlap_left_top = np.maximum(boxes0[..., :2], boxes1[..., :2])
    overlap_right_bottom = np.minimum(boxes0[..., 2:], boxes1[..., 2:])

    overlap_area = area_of(overlap_left_top, overlap_right_bottom)
    area0 = area_of(boxes0[..., :2], boxes0[..., 2:])
    area1 = area_of(boxes1[..., :2], boxes1[..., 2:])
    return overlap_area / (area0 + area1 - overlap_area + eps)


def area_of(left_top, right_bottom):
    """Compute the areas of rectangles given two corners.

    Args:
        left_top (N, 2): left top corner.
        right_bottom (N, 2): right bottom corner.

    Returns:
        area (N): return the area.
    """
    hw = np.clip(right_bottom - left_top, 0.0, None)
    return hw[..., 0] * hw[..., 1]


class NanoDetABC(metaclass=ABCMeta):
    def __init__(self, input_shape=[320, 320], reg_max=7, strides=[8, 16, 32], prob_threshold=0.4,
                 iou_threshold=0.3, num_candidate=1000, top_k=-1):
        self.strides = strides
        self.input_shape = input_shape
        self.reg_max = reg_max
        self.prob_threshold = prob_threshold
        self.iou_threshold = iou_threshold
        self.num_candidate = num_candidate
        self.top_k = top_k
        self.img_mean = [103.53, 116.28, 123.675]
        self.img_std = [57.375, 57.12, 58.395]
        self.input_size = (self.input_shape[1], self.input_shape[0])
        self.class_names = [
            'person', 'bicycle', 'car', 'motorcycle', 'airplane', 'bus',
            'train', 'truck', 'boat', 'traffic_light', 'fire_hydrant',
            'stop_sign', 'parking_meter', 'bench', 'bird', 'cat', 'dog',
            'horse', 'sheep', 'cow', 'elephant', 'bear', 'zebra', 'giraffe',
            'backpack', 'umbrella', 'handbag', 'tie', 'suitcase', 'frisbee',
            'skis', 'snowboard', 'sports_ball', 'kite', 'baseball_bat',
            'baseball_glove', 'skateboard', 'surfboard', 'tennis_racket',
            'bottle', 'wine_glass', 'cup', 'fork', 'knife', 'spoon', 'bowl',
            'banana', 'apple', 'sandwich', 'orange', 'broccoli', 'carrot',
            'hot_dog', 'pizza', 'donut', 'cake', 'chair', 'couch',
            'potted_plant', 'bed', 'dining_table', 'toilet', 'tv', 'laptop',
            'mouse', 'remote', 'keyboard', 'cell_phone', 'microwave',
            'oven', 'toaster', 'sink', 'refrigerator', 'book', 'clock',
            'vase', 'scissors', 'teddy_bear', 'hair_drier', 'toothbrush'
        ]

    def preprocess(self, img):
        # resize image
        ResizeM = get_resize_matrix((img.shape[1], img.shape[0]), self.input_size, True)
        img_resize = cv2.warpPerspective(img, ResizeM, dsize=self.input_size)
        # normalize image
        img_input = img_resize.astype(np.float32) / 255
        img_mean = np.array(self.img_mean, dtype=np.float32).reshape(1, 1, 3) / 255
        img_std = np.array(self.img_std, dtype=np.float32).reshape(1, 1, 3) / 255
        img_input = (img_input - img_mean) / img_std
        # expand dims
        img_input = np.transpose(img_input, [2, 0, 1])
        img_input = np.expand_dims(img_input, axis=0)
        return img_input, ResizeM

    def postprocess(self, scores, raw_boxes, ResizeM, raw_shape):
        # generate centers
        decode_boxes = []
        select_scores = []
        for stride, box_distribute, score in zip(self.strides, raw_boxes, scores):
            # centers
            fm_h = self.input_shape[0] / stride
            fm_w = self.input_shape[1] / stride
            h_range = np.arange(fm_h)
            w_range = np.arange(fm_w)
            ww, hh = np.meshgrid(w_range, h_range)
            ct_row = (hh.flatten() + 0.5) * stride
            ct_col = (ww.flatten() + 0.5) * stride
            center = np.stack((ct_col, ct_row, ct_col, ct_row), axis=1)

            # box distribution to distance
            reg_range = np.arange(self.reg_max + 1)
            box_distance = box_distribute.reshape((-1, self.reg_max + 1))
            box_distance = softmax(box_distance, axis=1)
            box_distance = box_distance * np.expand_dims(reg_range, axis=0)
            box_distance = np.sum(box_distance, axis=1).reshape((-1, 4))
            box_distance = box_distance * stride

            # top K candidate
            topk_idx = np.argsort(score.max(axis=1))[::-1]
            topk_idx = topk_idx[:self.num_candidate]
            center = center[topk_idx]
            score = score[topk_idx]
            box_distance = box_distance[topk_idx]

            # decode box
            decode_box = center + [-1, -1, 1, 1] * box_distance

            select_scores.append(score)
            decode_boxes.append(decode_box)

        # nms
        bboxes = np.concatenate(decode_boxes, axis=0)
        confidences = np.concatenate(select_scores, axis=0)
        picked_box_probs = []
        picked_labels = []
        for class_index in range(0, confidences.shape[1]):
            probs = confidences[:, class_index]
            mask = probs > self.prob_threshold
            probs = probs[mask]
            if probs.shape[0] == 0:
                continue
            subset_boxes = bboxes[mask, :]
            box_probs = np.concatenate([subset_boxes, probs.reshape(-1, 1)], axis=1)
            box_probs = hard_nms(
                box_probs,
                iou_threshold=self.iou_threshold,
                top_k=self.top_k,
            )
            picked_box_probs.append(box_probs)
            picked_labels.extend([class_index] * box_probs.shape[0])
        if not picked_box_probs:
            return np.array([]), np.array([]), np.array([])
        picked_box_probs = np.concatenate(picked_box_probs)

        # resize output boxes
        picked_box_probs[:, :4] = warp_boxes(
            picked_box_probs[:, :4], np.linalg.inv(ResizeM), raw_shape[1], raw_shape[0])
        return picked_box_probs[:, :4].astype(np.int32), np.array(picked_labels), picked_box_probs[:, 4]

    @abstractmethod
    def infer_image(self, img_input):
        pass

    def detect(self, img):
        raw_shape = img.shape
        img_input, ResizeM = self.preprocess(img)
        scores, raw_boxes = self.infer_image(img_input)
        bbox, label, score = self.postprocess(scores, raw_boxes, ResizeM, raw_shape)
        return bbox, label, score

    def draw_box(self, raw_img, bbox, label, score):
        img = raw_img.copy()
        all_box = [[x,]+y+[z,] for x,y,z in zip(label, bbox.tolist(), score)]
        img_draw = overlay_bbox_cv(img, all_box, self.class_names)
        return img_draw
    
    def detect_folder(self, img_fold, result_path):
        img_fold = Path(img_fold)
        result_path = Path(result_path)
        result_path.mkdir(parents=True, exist_ok=True)

        img_name_list = filter(
            lambda x: str(x).endswith('.png') or str(x).endswith('.jpg'),
            img_fold.iterdir()
        )
        img_name_list = list(img_name_list)
        print(f'find {len(img_name_list)} images')

        for img_path in tqdm(img_name_list):
            img = cv2.imread(str(img_path))
            bbox, label, score = self.detect(img)
            img_draw = self.draw_box(img, bbox, label, score)
            save_path = str(result_path / img_path.name.replace(".png", ".jpg"))
            cv2.imwrite(save_path, img_draw)


class NanoDetMNN(NanoDetABC):
    import MNN as MNNlib

    def __init__(self, model_path, *args, **kwargs):
        super(NanoDetMNN, self).__init__(*args, **kwargs)
        print(f'Using MNN as inference backend')
        print(f'Using weight: {model_path}')

        # load model
        self.model_path = model_path
        self.interpreter = self.MNNlib.Interpreter(self.model_path)
        self.session = self.interpreter.createSession()
        self.input_tensor = self.interpreter.getSessionInput(self.session)

    def infer_image(self, img_input):
        tmp_input = self.MNNlib.Tensor(
            (1, 3, self.input_size[1], self.input_size[0]),
            self.MNNlib.Halide_Type_Float,
            img_input,
            self.MNNlib.Tensor_DimensionType_Caffe
        )
        self.input_tensor.copyFrom(tmp_input)
        self.interpreter.runSession(self.session)
        score_out_name = ["792", "814", "836"]
        scores = [self.interpreter.getSessionOutput(self.session, x).getData() for x in score_out_name]
        scores = [np.reshape(x, (-1, 80)) for x in scores]
        boxes_out_name = ["795", "817", "839"]
        raw_boxes = [self.interpreter.getSessionOutput(self.session, x).getData() for x in boxes_out_name]
        raw_boxes = [np.reshape(x, (-1, 32)) for x in raw_boxes]
        return scores, raw_boxes


class NanoDetONNX(NanoDetABC):
    import onnxruntime as ort

    def __init__(self, model_path, *args, **kwargs):
        super(NanoDetONNX, self).__init__(*args, **kwargs)
        print(f'Using ONNX as inference backend')
        print(f'Using weight: {model_path}')

        # load model
        self.model_path = model_path
        self.ort_session = self.ort.InferenceSession(self.model_path)
        self.input_name = self.ort_session.get_inputs()[0].name

    def infer_image(self, img_input):
        inference_results = self.ort_session.run(None, {self.input_name: img_input})
        scores = [np.squeeze(x) for x in inference_results[:3]]
        raw_boxes = [np.squeeze(x) for x in inference_results[3:]]
        return scores, raw_boxes


class NanoDetTorch(NanoDetABC):
    import torch

    def __init__(self, model_path, cfg_path, *args, **kwargs):
        from nanodet.model.arch import build_model
        from nanodet.util import Logger, cfg, load_config, load_model_weight

        super(NanoDetTorch, self).__init__(*args, **kwargs)
        print(f'Using PyTorch as inference backend')
        print(f'Using weight: {model_path}')

        # load model
        self.model_path = model_path
        self.cfg_path = cfg_path
        load_config(cfg, cfg_path)
        self.logger = Logger(-1, cfg.save_dir, False)
        self.model = build_model(cfg.model)
        checkpoint = self.torch.load(model_path, map_location=lambda storage, loc: storage)
        load_model_weight(self.model, checkpoint, self.logger)

    def infer_image(self, img_input):
        self.model.train(False)
        with self.torch.no_grad():
            inference_results = self.model(self.torch.from_numpy(img_input))
        scores = [x.permute(0, 2, 3, 1).reshape((-1, 80)).sigmoid().detach().numpy() for x in inference_results[0]]
        raw_boxes = [x.permute(0, 2, 3, 1).reshape((-1, 32)).detach().numpy() for x in inference_results[1]]
        return scores, raw_boxes


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--model_path', dest='model_path', type=str, default='../model/nanodet-320.mnn')
    parser.add_argument('--cfg_path', dest='cfg_path', type=str, default='config/nanodet-m.yml')
    parser.add_argument('--img_fold', dest='img_fold', type=str, default='../imgs')
    parser.add_argument('--result_fold', dest='result_fold', type=str, default='./results')
    parser.add_argument('--input_shape', dest='input_shape', nargs=2, type=int, default=[320, 320])
    parser.add_argument('--backend', choices=['MNN', 'ONNX', 'torch'], default='MNN')
    args = parser.parse_args()

    print(f'Detecting {args.img_fold}')

    # load detector
    if args.backend == 'MNN':
        detector = NanoDetMNN(args.model_path, input_shape=args.input_shape)
    elif args.backend == 'ONNX':
        detector = NanoDetONNX(args.model_path, input_shape=args.input_shape)
    elif args.backend == 'torch':
        detector = NanoDetTorch(args.model_path, args.cfg_path, input_shape=args.input_shape)
    else:
        raise ValueError

    # detect folder
    detector.detect_folder(args.img_fold, args.result_fold)


def test_one():
    detector = NanoDetMNN('./weight/nanodet-320.mnn')
    img = cv2.imread('./data/2.jpg')
    bbox, label, score = detector.detect(img)
    img_draw = detector.draw_box(img, bbox, label, score)
    cv2.imshow('aa', img_draw[..., ::-1])
    cv2.waitKey(0)


if __name__ == '__main__':
    main()


from typing import Dict
import numpy as np
from alfred.utils.log import logger
import os
from abc import ABC, abstractclassmethod


class TensorInfo:
    def __init__(self, shape, dtype) -> None:
        self.shape = shape
        self.dtype = dtype


class WrapperBase:
    @abstractclassmethod
    def infer(self, inputs, return_dict=True):
        NotImplemented

    @abstractclassmethod
    def get_inputs_dict(self) -> Dict[str, TensorInfo]:
        NotImplemented

    @abstractclassmethod
    def get_outputs_dict(self) -> Dict[str, TensorInfo]:
        NotImplemented



import numpy as np
from alfred.utils.log import logger
import os

try:
    import MNN
except Exception as e:
    MNN = None

from .base import TensorInfo, WrapperBase


class MNNWrapper(WrapperBase):
    def __init__(self, onnx_f) -> None:
        """
        mnn not support convert in JIT
        """

        if os.path.basename(onnx_f).split(".")[-1] != "mnn":
            logger.error("model not mnn format, convert to mnn first.")
            exit()

        self.mnn_f = onnx_f
        logger.info(f"Loading from mnn model: {self.mnn_f}")

        # online parse onnx file to mnn format?
        self.interpreter = MNN.Interpreter(self.mnn_f)
        self.session = self.interpreter.createSession()
        self.input_tensor = self.interpreter.getSessionInput(self.session)
        self.output_tensor = self.interpreter.getSessionOutput(self.session)

    def np_as_mnn(self, inp_data):
        """
        inp data with NCHW
        """
        # logger.info(inp_data.shape)
        assert len(inp_data.shape) == 4, "mnn currently only support NCHW input."
        tmp_input = MNN.Tensor(
            inp_data.shape,
            MNN.Halide_Type_Float,
            inp_data,
            MNN.Tensor_DimensionType_Caffe,
        )
        return tmp_input

    def infer_si(self, inputs):
        if isinstance(inputs, list):
            inputs = np.array(inputs)

        tmp_input = self.np_as_mnn(inputs)
        self.input_tensor.copyFrom(tmp_input)

        self.interpreter.runSession(self.session)

        # output_tensor = self.interpreter.getSessionOutput(self.session, "output")
        output_tensor_dict = self.interpreter.getSessionOutputAll(self.session)

        res = {}
        for k, v in output_tensor_dict.items():
            tmp_output = MNN.Tensor(
                v.getShape(),
                v.getDataType(),
                np.ones(v.getShape()).astype(np.float32),
                MNN.Tensor_DimensionType_Caffe,
            )
            v.copyToHostTensor(tmp_output)
            res[k] = np.array(tmp_output.getData()).reshape(v.getShape())
        return res

    def infer(self, inputs):
        if isinstance(inputs, list):
            inputs = np.array(inputs)

        tmp_input = self.np_as_mnn(inputs)
        self.input_tensor.copyFrom(tmp_input)

        self.interpreter.runSession(self.session)

        # output_tensor = self.interpreter.getSessionOutput(self.session, "output")
        output_tensor_dict = self.interpreter.getSessionOutputAll(self.session)

        res = {}
        for k, v in output_tensor_dict.items():
            tmp_output = MNN.Tensor(
                v.getShape(),
                v.getDataType(),
                np.ones(v.getShape()).astype(np.float32),
                MNN.Tensor_DimensionType_Caffe,
            )
            v.copyToHostTensor(tmp_output)
            res[k] = np.array(tmp_output.getData()).reshape(v.getShape())
        return res

    def get_inputs_dict(self):
        input_tensor_dict = self.interpreter.getSessionInputAll(self.session)

        res = {}
        for k, v in input_tensor_dict.items():
            res[k] = TensorInfo(v.getShape(), v.getDataType())
        return res

    def get_outputs_dict(self):
        output_tensor_dict = self.interpreter.getSessionInputAll(self.session)

        res = {}
        for k, v in output_tensor_dict.items():
            res[k] = TensorInfo(v.getShape(), v.getDataType())
        return res



from wanwu.utils.image_utils import load_inputs_std_mean
from ..core.infer import Backends, BaseInfer
from ..core.modelloader import ModelLoader, TaskType
import time
import numpy as np
from .utils import box2cs, get_affine_transform, transform_preds_batch
from wanwu.base.base_task import TaskBase
import numpy as np
from wanwu import Det
from wanwu.track.deepsort import DeepSort
from alfred import logger
try:
    import cv2
except ImportError as e:
    cv2 = None


class Keypoints(BaseInfer):
    def __init__(
        self,
        onnx_f: str = None,
        backend: Backends = Backends.CPU_ORT,
        type: str = "alphapose_r50",
        batch_size=1,
        input_width=640,
        input_height=512,
        num_classes=80,
        fp16=False,
        timing=False,
        **kargs,
    ) -> None:
        super(Keypoints, self).__init__(
            type=type,
            onnx_f=onnx_f,
            backend=backend,
            batch_size=batch_size,
            input_width=input_width,
            input_height=input_height,
        )

        # add some CPU realtime Segmentation Models
        self.num_classes = num_classes

        self.timing = timing
        self.frame_time_cost = 0

        self.model_loader = ModelLoader(
            self.onnx_f,
            backend=self.backend,
            task_type=TaskType.KPS,
            fp16=fp16,
            num_classes=self.num_classes,
            batch_size=self.batch_size,
            input_width=self.input_width,
            input_height=self.input_height,
        )
        self.infer_wrapper = self.model_loader.get_infer_wrapper()
        self.pose_input_size = [self.input_height, self.input_width]

    def get_croped_images(self, boxes, ori_img):
        batch_images = []
        batch_centers = []
        batch_scales = []
        for box in boxes:
            c, s = box2cs(self.pose_input_size[::-1], box)
            batch_centers.append(c)
            batch_scales.append(s)
            # tt = time.perf_counter()
            trans = get_affine_transform(c, s, 0, self.pose_input_size[::-1])
            crop = cv2.warpAffine(
                ori_img, trans, self.pose_input_size[::-1], flags=cv2.INTER_LINEAR
            )
            # Alphapose using RGB as input
            crop = cv2.cvtColor(crop, cv2.COLOR_BGR2RGB)
            # ttt = time.perf_counter()
            # print(ttt -tt)
            crop = crop.astype(np.float32)
            batch_images.append(crop)
        return (
            np.array(batch_images).astype(np.float32),
            np.array(batch_centers),
            np.array(batch_scales),
        )

    def infer(self, raw_img, boxes):
        """
        this only for bottom up keypoints model which using boxes crop image patch first
        the send into keypoints detection model

        CAUTION: all keypoints model must wrap HWC=>CHW into model inside
        """
        if self.timing:
            t0 = time.perf_counter()
        batched_inputs, batch_centers, batch_scales = self.get_croped_images(
            boxes, raw_img
        )
        # print(batched_inputs.shape)
        res = self.infer_wrapper.infer(batched_inputs)
        if self.timing:
            t1 = time.perf_counter()
            self.frame_time_cost = round(t1 - t0, 3)
        # hardcoded on onnx model
        if isinstance(res, dict):
            cords = res["coords"]
            scores = res["maxvals"]
        else:
            # TensorRT returns list
            cords = res[1]
            scores = res[0]
            # print(cords.shape)
            # print(scores.shape)
        cords = self.postprocess(cords, batch_centers, batch_scales)
        return cords, scores

    def postprocess(self, outs, batch_centers, batch_scales):
        preds = transform_preds_batch(
            outs,
            batch_centers,
            batch_scales,
            (self.pose_input_size[1] // 4, self.pose_input_size[0] // 4),
        )
        return preds

    def load_inputs(self, img, std, mean, transpose=True, is_rgb=False):
        return load_inputs_std_mean(
            img,
            self.input_height,
            self.input_width,
            std=std,
            mean=mean,
            transpose=transpose,
            is_rgb=is_rgb,
        )

    def vis_res(self, img, res, class_names=None):
        print("Classes: ", res)


class TopDownKptsDetector(TaskBase):
    def __init__(self, kpts_model_f=None, track=False, **args) -> None:
        super(TopDownKptsDetector, self).__init__(**args)

        self.det = Det(
            type="yolox_tiny_coco",
            onnx_f=self.onnx_f,
            backend=self.backend,
            input_height=416,
            input_width=416,
            score_thr=0.45,
            fp16=False,
            timing=True,
        )
        self.kps = Keypoints(
            type="alp_halpe26_res50",
            onnx_f=kpts_model_f,
            backend=Backends.AUTO,
            input_height=256,
            input_width=192,
        )
        self.track = track
        if track:
            logger.info("Enabled tracking.")
            self.deepsort = DeepSort(
                Backends.AUTO, max_dist=0.4, max_iou_distance=0.8, max_age=20
            )

    def infer(self, im):
        inp = self.det.load_inputs(im, normalize_255=False, is_rgb=True)
        boxes, scores, labels = self.det.infer(inp, im.shape[0], im.shape[1])
        if self.track:
            boxes, track_ids, labels = self.deepsort.update(boxes, scores, labels, im)
        else:
            track_ids = None

        mask_idx = labels == 0
        boxes = boxes[mask_idx]

        if boxes.shape[0] > 0:
            kps_coords, score_kps = self.kps.infer(im, boxes)
        else:
            kps_coords = None
            score_kps = None
        return boxes, scores, labels, kps_coords, score_kps, track_ids



"""

We using a very simple feature extractor model
for appearance extraction, and using for matching
in do tracking.

It was barely based on DeepSort machanism.

"""
import numpy as np
from wanwu.core.enums import Backends

from .matching import NearestNeighborDistanceMetric
from .tracker import Tracker, Detection
from .reid import ReIDFeatureExtractor

__all__ = ["DeepSort"]


class DeepSort(object):
    def __init__(
        self,
        backend=Backends.AUTO,
        max_dist=0.2,
        max_iou_distance=0.7,
        max_age=70,
        n_init=3,
        nn_budget=100,
    ):

        self.extractor = ReIDFeatureExtractor(backend)
        self.height = 512
        self.width = 512

        max_cosine_distance = max_dist
        metric = NearestNeighborDistanceMetric(
            "euclidean", max_cosine_distance, nn_budget
        )
        self.tracker = Tracker(
            metric, max_iou_distance=max_iou_distance, max_age=max_age, n_init=n_init
        )

    def update(self, bbox_xyxy, confidences, classes, ori_img, use_yolo_preds=False):
        """
        boxes-in is xyxy order
        """
        if isinstance(bbox_xyxy, np.ndarray):
            bbox_xyxy = np.array(bbox_xyxy)
        bbox_xyxy = bbox_xyxy.astype(int)
        self.height, self.width = ori_img.shape[:2]
        # generate detections
        features = self._get_features(bbox_xyxy, ori_img)
        bbox_tlwh = self._xyxy_to_x1y1wh(bbox_xyxy)
        detections = [
            Detection(bbox_tlwh[i], conf, features[i])
            for i, conf in enumerate(confidences)
        ]

        # update tracker
        self.tracker.predict()
        self.tracker.update(detections, classes)

        # output bbox identities
        track_boxes = []
        track_ids = []
        track_labels = []
        for track in self.tracker.tracks:
            if not track.is_confirmed() or track.time_since_update > 1:
                continue
            box = track.to_tlwh()
            x1, y1, x2, y2 = self._tlwh_to_xyxy_single(box)
            track_id = track.track_id
            class_id = track.class_id

            track_boxes.append([x1, y1, x2, y2])
            track_ids.append(track_id)
            track_labels.append(class_id)

        if len(track_ids) > 0:
            track_boxes = np.stack(track_boxes, axis=0)
        return track_boxes, track_ids, track_labels

    def increment_ages(self):
        self.tracker.increment_ages()

    def _get_features(self, bbox_xyxy, ori_img):
        im_crops = []
        bbox_xyxy[:, :2] = bbox_xyxy[:, :2].clip(0)
        for box in bbox_xyxy:
            x1, y1, x2, y2 = box
            im = ori_img[y1:y2, x1:x2]
            im_crops.append(im)
        if im_crops:
            features = self.extractor.extract(im_crops)
        else:
            features = np.array([])
        return features

    def _tlwh_to_xyxy_single(self, bbox_tlwh):
        x, y, w, h = bbox_tlwh
        x1 = max(int(x), 0)
        x2 = min(int(x + w), self.width - 1)
        y1 = max(int(y), 0)
        y2 = min(int(y + h), self.height - 1)
        return x1, y1, x2, y2

    @staticmethod
    def _tlwh_to_xyxy(bbox_tlwh):
        if len(bbox_tlwh.shape) < 2:
            bbox_tlwh = np.expand_dims(bbox_tlwh, axis=0)
        t = bbox_tlwh[:, 0]
        l = bbox_tlwh[:, 1]
        bbox_tlwh[:, 2] = l + bbox_tlwh[:, 2]
        bbox_tlwh[:, 3] = t + bbox_tlwh[:, 3]
        bbox_tlwh[:, 0] = l
        bbox_tlwh[:, 1] = t
        return bbox_tlwh

    @staticmethod
    def _xyxy_to_x1y1wh(bbox_xyxy):
        if len(bbox_xyxy.shape) < 2:
            bbox_xyxy = np.expand_dims(bbox_xyxy, axis=0)
        x1 = bbox_xyxy[:, 0].clip(0)
        y1 = bbox_xyxy[:, 1].clip(0)
        bbox_xyxy[:, 2] = bbox_xyxy[:, 2] - x1
        bbox_xyxy[:, 3] = bbox_xyxy[:, 3] - y1
        return bbox_xyxy




import argparse
from wanwu.core.enums import Backends
from wanwu.det.det import Det
from alfred.utils.log import logger
import os
from wanwu.core.cellar import model_urls
from wanwu.core.modelloader import ModelLoader, TaskType
from .version import __version__


def arg_parse():
    """
    parse arguments
    :return:
    """
    parser = argparse.ArgumentParser(prog="wanwu")
    parser.add_argument(
        "--version", "-v", action="store_true", help="show version info."
    )

    main_sub_parser = parser.add_subparsers(dest="subparser_name")

    # =============== totrt part ================
    trt_parser = main_sub_parser.add_parser(
        "totrt", help="Convert model to trt with standared tasks."
    )
    trt_parser.add_argument("--model", "-m", help="onnx model path")
    trt_parser.add_argument(
        "--data_type", "-d", type=int, default=32, help="32, 16, 8 presents fp32, fp16, int8"
    )
    trt_parser.add_argument(
        "--batch_size", "-b", type=int, default=1, help="Batch size of the model."
    )
    trt_parser.add_argument(
        "--score_thr", "-s", type=float, default=0.2, help="score threshold of detection model, only work on det model."
    )
    trt_parser.add_argument(
        "--nms_thr", "-n", type=float, default=0.6, help="nms threshold of detection model, only work on det model."
    )
    trt_parser.add_argument(
        "--input_width", type=int, default=640, help="onnx model input width (for dynamic width onnx model)."
    )
    trt_parser.add_argument(
        "--input_height", type=int, default=512, help="onnx model input height (for dynamic height onnx model)."
    )
    trt_parser.add_argument("--min_shapes", nargs='+',
                            help="min_shapes of opt_params")
    trt_parser.add_argument("--opt_shapes", nargs='+',
                            help="opt_shapes of opt_params")
    trt_parser.add_argument("--max_shapes", nargs='+',
                            help="max_shapes of opt_params")
    trt_parser.add_argument(
        "--task", "-t", default='det', help="task name: det | kps | cls | seg"
    )
    trt_parser.add_argument(
        "--nms_plugin", default='efficient_nms', help="batched_nms | efficient_nms"
    )

    # # =============== check part ================
    # check_parser = main_sub_parser.add_parser(
    #     "ls", help="Check your onnx model is valid or not."
    # )
    # check_parser.add_argument("--model", "-m", help="onnx model path")
    # check_parser.add_argument("--print", action="store_true")
    return parser.parse_args()


class WanwuCli(object):
    def __init__(self):
        args = arg_parse()
        if args.version:
            print(__version__)
        else:
            if args.subparser_name == None:
                print("should provide at least one sub command, -h for detail.")
                exit(-1)
            self.model_path = args.model
            if args.model != None and os.path.exists(args.model):
                if args.subparser_name == "totrt":
                    self.totrt(self.model_path, args)
                elif args.subparser_name == "ls":
                    self.ls()

            else:
                print(
                    "{} does not exist or you should provide model path like `wanwu totrt -m model.onnx`.".format(
                        args.model
                    )
                )

    def ls(self):
        print('supported models" \n')
        print(model_urls)

    def search(self):
        pass

    def check(self, p):
        pass

    def totrt(self, onnx_f, args):
        bs = args.batch_size
        score_thr = args.score_thr
        nms_thr = args.nms_thr
        nms_plugin = args.nms_plugin
        t = args.task
        w = args.input_width
        h = args.input_height
        res = ModelLoader(onnx_f, backend=Backends.GPU_TENSORRT, task_type=TaskType.from_str(t),
                          batch_size=bs, score_thr=score_thr, input_height=h, input_width=w,
                          nms_thr=nms_thr, nms_plugin_name=nms_plugin)
        print('done. a new model should be generated into: {}'.format(res.onnx_f))


def main():
    WanwuCli()


if __name__ == "__main__":
    main()



import numpy as np
from PIL import Image

try:
    import torch
    import cv2
    from torch import Tensor
except ImportError as e:
    torch = None
    cv2 = None
    Tensor = None


def read_image_to_tensor(
    image: np.ndarray,
    is_half: bool = False,
) -> Tensor:
    """
    Parse an image to Tensor.

    Args:
        image (np.ndarray): the candidate ndarray image to be parsed to Tensor.
        is_half (bool): whether to transfer image to half. Default: False.
    """
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    image = np.ascontiguousarray(image, dtype=np.float32)  # uint8 to float32
    image = np.transpose(image / 255.0, [2, 0, 1])

    image = torch.from_numpy(image)
    image = image.half() if is_half else image.float()
    return image


def load_inputs(
    image, input_height, input_width, normalize_255=True, transpose=True, is_rgb=False
):
    if is_rgb:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    image = cv2.resize(image, (input_width, input_height))
    image = np.ascontiguousarray(image, dtype=np.float32)  # uint8 to float32
    if transpose:
        if normalize_255:
            image = np.transpose(image / 255.0, [2, 0, 1])
        else:
            image = np.transpose(image, [2, 0, 1])
    image = np.expand_dims(image, axis=0)
    return image


def load_inputs_naive(
    image, input_height, input_width, normalize_255=True, transpose=True
):
    """
    the image must be in RGB order
    """
    image = cv2.resize(image, (input_width, input_height))
    image = np.ascontiguousarray(image, dtype=np.float32)  # uint8 to float32
    if transpose:
        if normalize_255:
            image = np.transpose(image / 255.0, [2, 0, 1])
        else:
            image = np.transpose(image, [2, 0, 1])
    image = np.expand_dims(image, axis=0)
    return image


def load_inputs_std_mean(
    image, input_height, input_width, std=1, mean=0, transpose=True, is_rgb=False
):
    if is_rgb:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    image = cv2.resize(image, (input_width, input_height))
    image = np.ascontiguousarray(image, dtype=np.float32)  # uint8 to float32

    image = image.astype(np.float32)
    image -= mean
    image /= std
    if transpose:
        image = np.transpose(image, [2, 0, 1])
    image = np.expand_dims(image, axis=0)
    return image


def rescale_boxes(boxes, ori_h, ori_w, inp_h, inp_w):
    scale_x = ori_w * 1.0 / inp_w
    scale_y = ori_h * 1.0 / inp_h
    boxes[..., 0] *= scale_x
    boxes[..., 2] *= scale_x
    boxes[..., 1] *= scale_y
    boxes[..., 3] *= scale_y
    return boxes


def ensure_pil_bgr(image: Image.Image) -> np.ndarray:
    image_array = np.array(image)

    if image_array.shape[-1] == 3 and np.array_equal(
        image_array[..., :3], image_array[..., ::-1]
    ):
        return np.ascontiguousarray(image_array)

    bgr_array = image_array[..., ::-1]
    return np.ascontiguousarray(bgr_array)


def ensure_ndarray_bgr(image: np.ndarray) -> np.ndarray:
    if image.shape[-1] == 3 and np.array_equal(image[..., :3], image[..., ::-1]):
        return np.ascontiguousarray(image)

    bgr_array = image[..., ::-1]
    return np.ascontiguousarray(bgr_array)



from typing import Union
from wanwu.utils.boxes_utils import ltwh_to_x1y1x2y2
from wanwu.utils.image_utils import (
    ensure_ndarray_bgr,
    ensure_pil_bgr,
    rescale_boxes,
    load_inputs,
)
from ..core.infer import Backends, BaseInfer
from ..core.modelloader import ModelLoader, TaskType
import time
import numpy as np
import cv2
from PIL import Image


class Det(BaseInfer):
    def __init__(
        self,
        onnx_f: str = None,
        backend: Backends = Backends.CPU_ORT,
        type: str = "yolov5s_coco",
        batch_size=1,
        input_width=640,
        input_height=512,
        num_classes=80,
        score_thr=0.2,
        nms_thr=0.6,
        keep_to_topk=100,
        fp16=False,
        timing=False,
        do_nms=False,
        **kargs,
    ) -> None:
        super(Det, self).__init__(
            type=type,
            onnx_f=onnx_f,
            backend=backend,
            batch_size=batch_size,
            input_width=input_width,
            input_height=input_height,
        )

        self.num_classes = num_classes
        self.score_thr = score_thr
        self.nms_thr = nms_thr
        self.keep_to_topk = keep_to_topk
        self.do_nms = do_nms

        self.timing = timing
        self.frame_time_cost = 0

        model_loader = ModelLoader(
            self.onnx_f,
            backend=self.backend,
            task_type=TaskType.DET,
            fp16=fp16,
            num_classes=self.num_classes,
            batch_size=self.batch_size,
            input_width=self.input_width,
            input_height=self.input_height,
            score_thr=self.score_thr,
            nms_thr=self.nms_thr,
            keep_to_topk=self.keep_to_topk,
        )
        self.infer_wrapper = model_loader.get_infer_wrapper()

        self.crt_image_raw_w = 640
        self.crt_image_raw_h = 640

    def infer(self, inp, ori_h=None, ori_w=None):
        if ori_h is None or ori_w is None:
            ori_h, ori_w = self.crt_image_raw_h, self.crt_image_raw_w
        assert (
            inp.shape[-2] == self.input_width or inp.shape[-2] == self.input_height
        ), f"Seems your inp shape not equal to your Det input hw, please check. {inp.shape} vs {self.input_height}x{self.input_width}"
        if self.timing:
            t0 = time.perf_counter()
        res = self.infer_wrapper.infer(inp)
        if self.timing:
            t1 = time.perf_counter()
            self.frame_time_cost = round(t1 - t0, 3)

        boxes = None
        scores = None
        labels = None

        if self.backend == Backends.CPU_ORT:
            # solving ORT outputs
            if self.do_nms:
                # assue output like [1, 10647, 16]
                # [1, 25200, 21] for yolov7-face
                res = res["output"]
                boxes, scores, labels = self.do_nms_postprocess(res, ori_h, ori_w)
                # print(f"boxes: {boxes.shape}, scores: {scores.shape}, labels: {labels.shape}")
            else:
                if isinstance(res, dict):
                    # as boxes, scores, labels order
                    boxes = res["boxes"]
                    labels = res["labels"]
                    if "scores" in res.keys():
                        scores = res["scores"]
                    else:
                        scores = boxes[..., -1]
                        boxes = boxes[..., :-1]

                    shs = res["boxes"].shape
                    if len(shs) > 2:
                        n = shs[1]
                        if shs[0] == 1:
                            # if only one batch
                            boxes = boxes[0]
                            scores = scores[0]
                            labels = labels[0]

        elif self.backend == Backends.GPU_TENSORRT:
            """
            typically output of TensorRT should be:
            n: 1, 14
            boxes: 1, 100, 4
            scores: 1, 100
            labels: 1, 100
            """
            # solving TRT outputs
            shs = res[1].shape
            if len(shs) > 2:
                # TODO: add multi-batch supported here.
                n = np.squeeze(res[0], axis=-1)[0]
                if shs[0] == 1:
                    boxes = res[1][0][:n]
                    scores = res[2][0][:n]
                    labels = res[3][0][:n]
                else:
                    boxes = res[1][:n]
                    scores = res[2][:n]
                    labels = res[3][:n]
            else:
                n = res[0]
                boxes = res[1][:n]
                scores = res[2][:n]
                labels = res[3][:n]
        else:
            ValueError(f"Backend {self.backend} not implemented for Det infer")

        # finally get boxes, scores, labels
        boxes, scores, labels = self.postprocess(ori_h, ori_w, boxes, scores, labels)
        return boxes, scores, labels

    def load_inputs(self, im, normalize_255=False, is_rgb=True):
        self.crt_image_raw_h = im.shape[0]
        self.crt_image_raw_w = im.shape[1]
        return load_inputs(
            im,
            self.input_height,
            self.input_width,
            normalize_255=normalize_255,
            is_rgb=is_rgb,
        )

    def do_nms_postprocess(self, outs, ori_h, ori_w, type="yolo-face"):
        confidences = []
        boxes = []
        class_ids = []
        landmarks = []
        ratioh, ratiow = ori_h / self.input_height, ori_w / self.input_width
        if len(outs.shape) > 2:
            # get batch out
            outs = outs[0]
        # we might have different type for post-process
        if type == "yolo-face":
            for detection in outs:
                if len(detection) == 21:
                    # yolov7 face
                    # 21 corresponds to: 1 channel, 4 bounding box, 1 score, 15 landmarks
                    confidence = detection[5]
                else:
                    confidence = detection[15]
                # if confidence > self.confThreshold and detection[4] > self.objThreshold:
                # if float(confidence) > self.score_thr:
                # this objness
                if detection[4] > 0.5:
                    cx = detection[0]
                    cy = detection[1]
                    w = detection[2]
                    h = detection[3]
                    left = int(cx - w / 2)
                    top = int(cy - h / 2)

                    confidences.append(float(confidence))
                    boxes.append([left, top, w, h])
                    class_ids.append(0)
                    landmark = detection[5:15] * np.tile(
                        np.float32([ratiow, ratioh]), 5
                    )
                    landmarks.append(landmark.astype(np.int32))
        else:
            raise ValueError(f"{type} not supported now, only yolo-face support")

        indices = cv2.dnn.NMSBoxes(boxes, confidences, self.score_thr, self.nms_thr)
        boxes = np.array(boxes)
        confidences = np.array(confidences)
        class_ids = np.array(class_ids)

        boxes = boxes[indices]
        class_ids = class_ids[indices]
        confidences = confidences[indices]
        if len(boxes) > 0:
            boxes = ltwh_to_x1y1x2y2(boxes)
        return boxes, confidences, class_ids

    def postprocess(self, ori_h, ori_w, boxes, scores, labels):
        """
        rescale box,
        and return final:
        boxes, scores, labels in order
        """
        if len(boxes) > 0:
            boxes = rescale_boxes(
                boxes, ori_h, ori_w, self.input_height, self.input_width
            )
            # filter again on score
            idx = scores > self.score_thr
            boxes = boxes[idx]
            scores = scores[idx]
            labels = labels[idx]
        return boxes, scores, labels

    def vis_res(
        self,
        img: Union[np.ndarray, Image.Image],
        bboxes,
        scores,
        labels=None,
        track_ids=None,
        class_names=None,
        transparent=False,
    ):
        from alfred.vis.image.det import visualize_det_cv2_part

        if isinstance(img, Image.Image):
            img = ensure_pil_bgr(img)
        else:
            # np.ndarray also need ensure bgr for cv2 to visualize
            img = ensure_ndarray_bgr(img)

        img = visualize_det_cv2_part(
            img,
            scores,
            labels,
            bboxes,
            thresh=self.score_thr,
            track_ids=track_ids,
            class_names=class_names,
            font_scale=0.8,
            line_thickness=2,
            show_time=self.frame_time_cost,
            transparent=transparent,
        )
        return img



from wanwu.cls.cls import Classify
from wanwu.core.cellar import model_urls
from wanwu.core.enums import Backends
import numpy as np


class ReIDFeatureExtractor:
    def __init__(self, backend: Backends) -> None:
        self.model = Classify(
            type="reid_osnet_x0_25", backend=backend, input_height=256, input_width=128
        )

    def extract(self, imgs):
        """
        imgs should be NDArray or list of arrays

        ONNX model already solves permute and normalization
        """
        imgs_input = []
        for im in imgs:
            im = self.model.load_inputs(im, transpose=False, is_rgb=True)
            imgs_input.append(im)
        feas = self.model.infer(imgs_input)
        if isinstance(feas, dict):
            feas = feas["feas"]
        else:
            feas = feas[0]
            if len(feas.shape) > 2:
                feas = np.squeeze(feas, axis=1)
        return feas



from yolort.runtime.yolo_graphsurgeon import YOLOGraphSurgeon
import sys

checkpoint_path = sys.argv[1]

yolo_gs = YOLOGraphSurgeon(
    checkpoint_path,
    version="r6.0",
    enable_dynamic=False,
)

yolo_gs.register_nms()

# Export the ONNX model
yolo_gs.save('weights/a.onnx')


# vim: expandtab:ts=4:sw=4
from __future__ import absolute_import
import numpy as np
from . import kalman_filter
from . import linear_assignment
from . import matching as iou_matching


class Detection(object):
    """
    This class represents a bounding box detection in a single image.

    Parameters
    ----------
    tlwh : array_like
        Bounding box in format `(x, y, w, h)`.
    confidence : float
        Detector confidence score.
    feature : array_like
        A feature vector that describes the object contained in this image.

    Attributes
    ----------
    tlwh : ndarray
        Bounding box in format `(top left x, top left y, width, height)`.
    confidence : ndarray
        Detector confidence score.
    feature : ndarray | NoneType
        A feature vector that describes the object contained in this image.

    """

    def __init__(self, tlwh, confidence, feature):
        self.tlwh = np.asarray(tlwh, dtype=np.float)
        self.confidence = float(confidence)
        self.feature = np.asarray(feature, dtype=np.float32)

    def to_tlbr(self):
        """Convert bounding box to format `(min x, min y, max x, max y)`, i.e.,
        `(top left, bottom right)`.
        """
        ret = self.tlwh.copy()
        ret[2:] += ret[:2]
        return ret

    def to_xyah(self):
        """Convert bounding box to format `(center x, center y, aspect ratio,
        height)`, where the aspect ratio is `width / height`.
        """
        ret = self.tlwh.copy()
        ret[:2] += ret[2:] / 2
        ret[2] /= ret[3]
        return ret


class TrackState:
    """
    Enumeration type for the single target track state. Newly created tracks are
    classified as `tentative` until enough evidence has been collected. Then,
    the track state is changed to `confirmed`. Tracks that are no longer alive
    are classified as `deleted` to mark them for removal from the set of active
    tracks.

    """

    Tentative = 1
    Confirmed = 2
    Deleted = 3


class Tracklet:
    """
    A single target track with state space `(x, y, a, h)` and associated
    velocities, where `(x, y)` is the center of the bounding box, `a` is the
    aspect ratio and `h` is the height.

    Parameters
    ----------
    mean : ndarray
        Mean vector of the initial state distribution.
    covariance : ndarray
        Covariance matrix of the initial state distribution.
    track_id : int
        A unique track identifier.
    n_init : int
        Number of consecutive detections before the track is confirmed. The
        track state is set to `Deleted` if a miss occurs within the first
        `n_init` frames.
    max_age : int
        The maximum number of consecutive misses before the track state is
        set to `Deleted`.
    feature : Optional[ndarray]
        Feature vector of the detection this track originates from. If not None,
        this feature is added to the `features` cache.

    Attributes
    ----------
    mean : ndarray
        Mean vector of the initial state distribution.
    covariance : ndarray
        Covariance matrix of the initial state distribution.
    track_id : int
        A unique track identifier.
    hits : int
        Total number of measurement updates.
    age : int
        Total number of frames since first occurance.
    time_since_update : int
        Total number of frames since last measurement update.
    state : TrackState
        The current track state.
    features : List[ndarray]
        A cache of features. On each measurement update, the associated feature
        vector is added to this list.

    """

    def __init__(
        self, mean, covariance, track_id, class_id, n_init, max_age, feature=None
    ):
        self.mean = mean
        self.covariance = covariance
        self.track_id = track_id
        self.class_id = class_id
        self.hits = 1
        self.age = 1
        self.time_since_update = 0
        self.yolo_bbox = [0, 0, 0, 0]

        self.state = TrackState.Tentative
        self.features = []
        if feature is not None:
            self.features.append(feature)

        self._n_init = n_init
        self._max_age = max_age

    def to_tlwh(self):
        ret = self.mean[:4].copy()
        ret[2] *= ret[3]
        ret[:2] -= ret[2:] / 2
        return ret
    
    def to_xyxy(self):
        ret = self.mean[:4].copy()
        pass 

    def to_tlbr(self):
        """Get kf estimated current position in bounding box format `(min x, miny, max x,
        max y)`.

        Returns
        -------
        ndarray
            The predicted kf bounding box.

        """
        ret = self.to_tlwh()
        ret[2:] = ret[:2] + ret[2:]
        return ret

    def get_yolo_pred(self):
        """Get yolo prediction`.

        Returns
        -------
        ndarray
            The yolo bounding box.

        """
        return self.yolo_bbox

    def increment_age(self):
        self.age += 1
        self.time_since_update += 1

    def predict(self, kf):
        """Propagate the state distribution to the current time step using a
        Kalman filter prediction step.

        Parameters
        ----------
        kf : kalman_filter.KalmanFilter
            The Kalman filter.

        """
        self.mean, self.covariance = kf.predict(self.mean, self.covariance)
        self.increment_age()

    def update(self, kf, detection, class_id):
        """Perform Kalman filter measurement update step and update the feature
        cache.

        Parameters
        ----------
        kf : kalman_filter.KalmanFilter
            The Kalman filter.
        detection : Detection
            The associated detection.

        """
        self.yolo_bbox = detection
        self.mean, self.covariance = kf.update(
            self.mean, self.covariance, detection.to_xyah()
        )
        self.features.append(detection.feature)
        self.class_id = class_id

        self.hits += 1
        self.time_since_update = 0
        if self.state == TrackState.Tentative and self.hits >= self._n_init:
            self.state = TrackState.Confirmed

    def mark_missed(self):
        """Mark this track as missed (no association at the current time step)."""
        if self.state == TrackState.Tentative:
            self.state = TrackState.Deleted
        elif self.time_since_update > self._max_age:
            self.state = TrackState.Deleted

    def is_tentative(self):
        """Returns True if this track is tentative (unconfirmed)."""
        return self.state == TrackState.Tentative

    def is_confirmed(self):
        """Returns True if this track is confirmed."""
        return self.state == TrackState.Confirmed

    def is_deleted(self):
        """Returns True if this track is dead and should be deleted."""
        return self.state == TrackState.Deleted


class Tracker:
    """
    This is the multi-target tracker.
    Parameters
    ----------
    metric : nn_matching.NearestNeighborDistanceMetric
        A distance metric for measurement-to-track association.
    max_age : int
        Maximum number of missed misses before a track is deleted.
    n_init : int
        Number of consecutive detections before the track is confirmed. The
        track state is set to `Deleted` if a miss occurs within the first
        `n_init` frames.
    Attributes
    ----------
    metric : nn_matching.NearestNeighborDistanceMetric
        The distance metric used for measurement to track association.
    max_age : int
        Maximum number of missed misses before a track is deleted.
    n_init : int
        Number of frames that a track remains in initialization phase.
    kf : kalman_filter.KalmanFilter
        A Kalman filter to filter target trajectories in image space.
    tracks : List[Track]
        The list of active tracks at the current time step.
    """

    GATING_THRESHOLD = np.sqrt(kalman_filter.chi2inv95[4])

    def __init__(self, metric, max_iou_distance=0.9, max_age=30, n_init=3, _lambda=0):
        self.metric = metric
        self.max_iou_distance = max_iou_distance
        self.max_age = max_age
        self.n_init = n_init
        self._lambda = _lambda

        self.kf = kalman_filter.KalmanFilter()
        self.tracks = []
        self._next_id = 1

    def predict(self):
        """Propagate track state distributions one time step forward.

        This function should be called once every time step, before `update`.
        """
        for track in self.tracks:
            track.predict(self.kf)

    def increment_ages(self):
        for track in self.tracks:
            track.increment_age()
            track.mark_missed()

    def update(self, detections, classes):
        """Perform measurement update and track management.

        Parameters
        ----------
        detections : List[deep_sort.detection.Detection]
            A list of detections at the current time step.

        """
        # Run matching cascade.
        matches, unmatched_tracks, unmatched_detections = self._match(detections)

        # Update track set.
        for track_idx, detection_idx in matches:
            self.tracks[track_idx].update(
                self.kf, detections[detection_idx], classes[detection_idx]
            )
        for track_idx in unmatched_tracks:
            self.tracks[track_idx].mark_missed()
        for detection_idx in unmatched_detections:
            self._initiate_track(
                detections[detection_idx], classes[detection_idx].item()
            )
        self.tracks = [t for t in self.tracks if not t.is_deleted()]

        # Update distance metric.
        active_targets = [t.track_id for t in self.tracks if t.is_confirmed()]
        features, targets = [], []
        for track in self.tracks:
            if not track.is_confirmed():
                continue
            features += track.features
            targets += [track.track_id for _ in track.features]
            track.features = []
        self.metric.partial_fit(
            np.asarray(features), np.asarray(targets), active_targets
        )

    def _full_cost_metric(self, tracks, dets, track_indices, detection_indices):
        """
        This implements the full lambda-based cost-metric. However, in doing so, it disregards
        the possibility to gate the position only which is provided by
        linear_assignment.gate_cost_matrix(). Instead, I gate by everything.
        Note that the Mahalanobis distance is itself an unnormalised metric. Given the cosine
        distance being normalised, we employ a quick and dirty normalisation based on the
        threshold: that is, we divide the positional-cost by the gating threshold, thus ensuring
        that the valid values range 0-1.
        Note also that the authors work with the squared distance. I also sqrt this, so that it
        is more intuitive in terms of values.
        """
        # Compute First the Position-based Cost Matrix
        pos_cost = np.empty([len(track_indices), len(detection_indices)])
        msrs = np.asarray([dets[i].to_xyah() for i in detection_indices])
        for row, track_idx in enumerate(track_indices):
            pos_cost[row, :] = (
                np.sqrt(
                    self.kf.gating_distance(
                        tracks[track_idx].mean,
                        tracks[track_idx].covariance,
                        msrs,
                        False,
                    )
                )
                / self.GATING_THRESHOLD
            )
        pos_gate = pos_cost > 1.0
        # Now Compute the Appearance-based Cost Matrix
        app_cost = self.metric.distance(
            np.array([dets[i].feature for i in detection_indices]),
            np.array([tracks[i].track_id for i in track_indices]),
        )
        app_gate = app_cost > self.metric.matching_threshold
        # Now combine and threshold
        cost_matrix = self._lambda * pos_cost + (1 - self._lambda) * app_cost
        cost_matrix[np.logical_or(pos_gate, app_gate)] = linear_assignment.INFTY_COST
        # Return Matrix
        return cost_matrix

    def _match(self, detections):
        # Split track set into confirmed and unconfirmed tracks.
        confirmed_tracks = [i for i, t in enumerate(self.tracks) if t.is_confirmed()]
        unconfirmed_tracks = [
            i for i, t in enumerate(self.tracks) if not t.is_confirmed()
        ]

        # Associate confirmed tracks using appearance features.
        (
            matches_a,
            unmatched_tracks_a,
            unmatched_detections,
        ) = linear_assignment.matching_cascade(
            self._full_cost_metric,
            linear_assignment.INFTY_COST
            - 1,  # no need for self.metric.matching_threshold here,
            self.max_age,
            self.tracks,
            detections,
            confirmed_tracks,
        )

        # Associate remaining tracks together with unconfirmed tracks using IOU.
        iou_track_candidates = unconfirmed_tracks + [
            k for k in unmatched_tracks_a if self.tracks[k].time_since_update == 1
        ]
        unmatched_tracks_a = [
            k for k in unmatched_tracks_a if self.tracks[k].time_since_update != 1
        ]
        (
            matches_b,
            unmatched_tracks_b,
            unmatched_detections,
        ) = linear_assignment.min_cost_matching(
            iou_matching.iou_cost,
            self.max_iou_distance,
            self.tracks,
            detections,
            iou_track_candidates,
            unmatched_detections,
        )

        matches = matches_a + matches_b
        unmatched_tracks = list(set(unmatched_tracks_a + unmatched_tracks_b))
        return matches, unmatched_tracks, unmatched_detections

    def _initiate_track(self, detection, class_id):
        mean, covariance = self.kf.initiate(detection.to_xyah())
        self.tracks.append(
            Tracklet(
                mean,
                covariance,
                self._next_id,
                class_id,
                self.n_init,
                self.max_age,
                detection.feature,
            )
        )
        self._next_id += 1



# vim: expandtab:ts=4:sw=4
import numpy as np
from . import linear_assignment


def iou(bbox, candidates):
    """Computer intersection over union.

    Parameters
    ----------
    bbox : ndarray
        A bounding box in format `(top left x, top left y, width, height)`.
    candidates : ndarray
        A matrix of candidate bounding boxes (one per row) in the same format
        as `bbox`.

    Returns
    -------
    ndarray
        The intersection over union in [0, 1] between the `bbox` and each
        candidate. A higher score means a larger fraction of the `bbox` is
        occluded by the candidate.

    """
    bbox_tl, bbox_br = bbox[:2], bbox[:2] + bbox[2:]
    candidates_tl = candidates[:, :2]
    candidates_br = candidates[:, :2] + candidates[:, 2:]

    tl = np.c_[
        np.maximum(bbox_tl[0], candidates_tl[:, 0])[:, np.newaxis],
        np.maximum(bbox_tl[1], candidates_tl[:, 1])[:, np.newaxis],
    ]
    br = np.c_[
        np.minimum(bbox_br[0], candidates_br[:, 0])[:, np.newaxis],
        np.minimum(bbox_br[1], candidates_br[:, 1])[:, np.newaxis],
    ]
    wh = np.maximum(0.0, br - tl)

    area_intersection = wh.prod(axis=1)
    area_bbox = bbox[2:].prod()
    area_candidates = candidates[:, 2:].prod(axis=1)
    return area_intersection / (area_bbox + area_candidates - area_intersection)


def iou_cost(tracks, detections, track_indices=None, detection_indices=None):
    """An intersection over union distance metric.

    Parameters
    ----------
    tracks : List[deep_sort.track.Track]
        A list of tracks.
    detections : List[deep_sort.detection.Detection]
        A list of detections.
    track_indices : Optional[List[int]]
        A list of indices to tracks that should be matched. Defaults to
        all `tracks`.
    detection_indices : Optional[List[int]]
        A list of indices to detections that should be matched. Defaults
        to all `detections`.

    Returns
    -------
    ndarray
        Returns a cost matrix of shape
        len(track_indices), len(detection_indices) where entry (i, j) is
        `1 - iou(tracks[track_indices[i]], detections[detection_indices[j]])`.

    """
    if track_indices is None:
        track_indices = np.arange(len(tracks))
    if detection_indices is None:
        detection_indices = np.arange(len(detections))

    cost_matrix = np.zeros((len(track_indices), len(detection_indices)))
    for row, track_idx in enumerate(track_indices):
        if tracks[track_idx].time_since_update > 1:
            cost_matrix[row, :] = linear_assignment.INFTY_COST
            continue

        bbox = tracks[track_idx].to_tlwh()
        candidates = np.asarray([detections[i].tlwh for i in detection_indices])
        cost_matrix[row, :] = 1.0 - iou(bbox, candidates)
    return cost_matrix


def _pdist(a, b):
    """Compute pair-wise squared distance between points in `a` and `b`.
    Parameters
    ----------
    a : array_like
        An NxM matrix of N samples of dimensionality M.
    b : array_like
        An LxM matrix of L samples of dimensionality M.
    Returns
    -------
    ndarray
        Returns a matrix of size len(a), len(b) such that eleement (i, j)
        contains the squared distance between `a[i]` and `b[j]`.
    """
    a, b = np.asarray(a), np.asarray(b)
    if len(a) == 0 or len(b) == 0:
        return np.zeros((len(a), len(b)))
    a2, b2 = np.square(a).sum(axis=1), np.square(b).sum(axis=1)
    r2 = -2.0 * np.dot(a, b.T) + a2[:, None] + b2[None, :]
    r2 = np.clip(r2, 0.0, float(np.inf))
    return r2


def _cosine_distance(a, b, data_is_normalized=False):
    """Compute pair-wise cosine distance between points in `a` and `b`.
    Parameters
    ----------
    a : array_like
        An NxM matrix of N samples of dimensionality M.
    b : array_like
        An LxM matrix of L samples of dimensionality M.
    data_is_normalized : Optional[bool]
        If True, assumes rows in a and b are unit length vectors.
        Otherwise, a and b are explicitly normalized to lenght 1.
    Returns
    -------
    ndarray
        Returns a matrix of size len(a), len(b) such that eleement (i, j)
        contains the squared distance between `a[i]` and `b[j]`.
    """
    if not data_is_normalized:
        a = np.asarray(a) / np.linalg.norm(a, axis=1, keepdims=True)
        b = np.asarray(b) / np.linalg.norm(b, axis=1, keepdims=True)
    return 1.0 - np.dot(a, b.T)


def _nn_euclidean_distance(x, y):
    """Helper function for nearest neighbor distance metric (Euclidean).
    Parameters
    ----------
    x : ndarray
        A matrix of N row-vectors (sample points).
    y : ndarray
        A matrix of M row-vectors (query points).
    Returns
    -------
    ndarray
        A vector of length M that contains for each entry in `y` the
        smallest Euclidean distance to a sample in `x`.
    """
    distances = _pdist(x, y)
    return np.maximum(0.0, distances.min(axis=0))


def _nn_cosine_distance(x, y):
    """Helper function for nearest neighbor distance metric (cosine).
    Parameters
    ----------
    x : ndarray
        A matrix of N row-vectors (sample points).
    y : ndarray
        A matrix of M row-vectors (query points).
    Returns
    -------
    ndarray
        A vector of length M that contains for each entry in `y` the
        smallest cosine distance to a sample in `x`.
    """
    distances = _cosine_distance(x, y)
    return distances.min(axis=0)


class NearestNeighborDistanceMetric(object):
    """
    A nearest neighbor distance metric that, for each target, returns
    the closest distance to any sample that has been observed so far.
    Parameters
    ----------
    metric : str
        Either "euclidean" or "cosine".
    matching_threshold: float
        The matching threshold. Samples with larger distance are considered an
        invalid match.
    budget : Optional[int]
        If not None, fix samples per class to at most this number. Removes
        the oldest samples when the budget is reached.
    Attributes
    ----------
    samples : Dict[int -> List[ndarray]]
        A dictionary that maps from target identities to the list of samples
        that have been observed so far.
    """

    def __init__(self, metric, matching_threshold, budget=None):
        if metric == "euclidean":
            self._metric = _nn_euclidean_distance
        elif metric == "cosine":
            self._metric = _nn_cosine_distance
        else:
            raise ValueError("Invalid metric; must be either 'euclidean' or 'cosine'")
        self.matching_threshold = matching_threshold
        self.budget = budget
        self.samples = {}

    def partial_fit(self, features, targets, active_targets):
        """Update the distance metric with new data.
        Parameters
        ----------
        features : ndarray
            An NxM matrix of N features of dimensionality M.
        targets : ndarray
            An integer array of associated target identities.
        active_targets : List[int]
            A list of targets that are currently present in the scene.
        """
        for feature, target in zip(features, targets):
            self.samples.setdefault(target, []).append(feature)
            if self.budget is not None:
                self.samples[target] = self.samples[target][-self.budget :]
        self.samples = {k: self.samples[k] for k in active_targets}

    def distance(self, features, targets):
        """Compute distance between features and targets.
        Parameters
        ----------
        features : ndarray
            An NxM matrix of N features of dimensionality M.
        targets : List[int]
            A list of targets to match the given `features` against.
        Returns
        -------
        ndarray
            Returns a cost matrix of shape len(targets), len(features), where
            element (i, j) contains the closest squared distance between
            `targets[i]` and `features[j]`.
        """
        cost_matrix = np.zeros((len(targets), len(features)))
        for i, target in enumerate(targets):
            cost_matrix[i, :] = self._metric(self.samples[target], features)
        return cost_matrix



import numpy as np
try:
    import cv2
except ImportError as e:
    cv2 = None
    

def affine_transform(pt, trans_mat):
    """Apply an affine transformation to the points.
    Args:
        pt (np.ndarray): a 2 dimensional point to be transformed
        trans_mat (np.ndarray): 2x3 matrix of an affine transform
    Returns:
        np.ndarray: Transformed points.
    """
    assert len(pt) == 2
    new_pt = np.array(trans_mat) @ np.array([pt[0], pt[1], 1.0])

    return new_pt


def _get_3rd_point(a, b):
    """To calculate the affine matrix, three pairs of points are required. This
    function is used to get the 3rd point, given 2D points a & b.
    The 3rd point is defined by rotating vector `a - b` by 90 degrees
    anticlockwise, using b as the rotation center.
    Args:
        a (np.ndarray): point(x,y)
        b (np.ndarray): point(x,y)
    Returns:
        np.ndarray: The 3rd point.
    """
    assert len(a) == 2
    assert len(b) == 2
    direction = a - b
    third_pt = b + np.array([-direction[1], direction[0]], dtype=np.float32)

    return third_pt


def get_affine_transform(center, scale, rot, output_size, shift=(0.0, 0.0), inv=False):
    """Get the affine transform matrix, given the center/scale/rot/output_size.
    Args:
        center (np.ndarray[2, ]): Center of the bounding box (x, y).
        scale (np.ndarray[2, ]): Scale of the bounding box
            wrt [width, height].
        rot (float): Rotation angle (degree).
        output_size (np.ndarray[2, ] | list(2,)): Size of the
            destination heatmaps.
        shift (0-100%): Shift translation ratio wrt the width/height.
            Default (0., 0.).
        inv (bool): Option to inverse the affine transform direction.
            (inv=False: src->dst or inv=True: dst->src)
    Returns:
        np.ndarray: The transform matrix.
    """
    assert len(center) == 2
    assert len(scale) == 2
    assert len(output_size) == 2
    assert len(shift) == 2

    # pixel_std is 200.
    scale_tmp = scale * 200.0

    shift = np.array(shift)
    src_w = scale_tmp[0]
    dst_w = output_size[0]
    dst_h = output_size[1]

    rot_rad = np.pi * rot / 180
    src_dir = rotate_point([0.0, src_w * -0.5], rot_rad)
    dst_dir = np.array([0.0, dst_w * -0.5])

    src = np.zeros((3, 2), dtype=np.float32)
    src[0, :] = center + scale_tmp * shift
    src[1, :] = center + src_dir + scale_tmp * shift
    src[2, :] = _get_3rd_point(src[0, :], src[1, :])

    dst = np.zeros((3, 2), dtype=np.float32)
    dst[0, :] = [dst_w * 0.5, dst_h * 0.5]
    dst[1, :] = np.array([dst_w * 0.5, dst_h * 0.5]) + dst_dir
    dst[2, :] = _get_3rd_point(dst[0, :], dst[1, :])

    if inv:
        trans = cv2.getAffineTransform(np.float32(dst), np.float32(src))
    else:
        trans = cv2.getAffineTransform(np.float32(src), np.float32(dst))
    return trans


def rotate_point(pt, angle_rad):
    """Rotate a point by an angle.
    Args:
        pt (list[float]): 2 dimensional point to be rotated
        angle_rad (float): rotation angle by radian
    Returns:
        list[float]: Rotated point.
    """
    assert len(pt) == 2
    sn, cs = np.sin(angle_rad), np.cos(angle_rad)
    new_x = pt[0] * cs - pt[1] * sn
    new_y = pt[0] * sn + pt[1] * cs
    rotated_pt = [new_x, new_y]

    return rotated_pt


def box2cs(input_size, box, box_format="xyxy"):
    """This encodes bbox(x,y,w,h) into (center, scale)

    ori_image_size means original large image size, which box coordinates according to

    Args:
        x, y, w, h

    Returns:
        tuple: A tuple containing center and scale.

        - np.ndarray[float32](2,): Center of the bbox (x, y).
        - np.ndarray[float32](2,): Scale of the bbox w & h.
    """
    if box_format == "xyxy":
        x, y, x2, y2 = box
        w = x2 - x
        h = y2 - y
    elif box_format == "xywh":
        x, y, w, h = box[:4]
    elif box_format == "cxcywh":
        cx, cy, w, h = box[:4]
        x = cx - w * 0.5
        y = cy - h * 0.5
    aspect_ratio = input_size[0] / input_size[1]
    center = np.array([x + w * 0.5, y + h * 0.5], dtype=np.float32)

    if w > aspect_ratio * h:
        h = w * 1.0 / aspect_ratio
    elif w < aspect_ratio * h:
        w = h * aspect_ratio

    # pixel std is 200.0
    scale = np.array([w / 200.0, h / 200.0], dtype=np.float32)
    scale = scale * 1.25
    return center, scale


def transform_preds_batch(coords, center, scale, output_size, use_udp=False):
    """
    coords: [B, 17, 2]
    """
    # assert coords.shape[-1] in (2, 4, 5)
    assert len(output_size) == 2

    # Recover the scale which is normalized by a factor of 200.
    scale = scale * 200.0

    if use_udp:
        scale_x = scale[:, 0] / (output_size[0] - 1.0)
        scale_y = scale[:, 1] / (output_size[1] - 1.0)
    else:
        scale_x = scale[:, 0] / output_size[0]
        scale_y = scale[:, 1] / output_size[1]

    target_coords = np.ones_like(coords)
    for i in range(coords.shape[0]):
        c = center[i]
        s = scale[i]

        target_coords[i, :, 0] = coords[i, :, 0] * scale_x[i] + c[0] - s[0] * 0.5
        target_coords[i, :, 1] = coords[i, :, 1] * scale_y[i] + c[1] - s[1] * 0.5
    return target_coords



import cv2
import numpy as np
from openvino.inference_engine import IECore
from openvino.inference_engine import IENetwork
from blazeface import BlazeFace


def load_to_IE(model):
    # Loading the Inference Engine API
    ie = IECore()
    # Loading IR files
    # net = IENetwork(model=model + ".xml", weights=model + ".bin")
    net = ie.read_network(model)
    # Loading the network to the inference engine
    exec_net = ie.load_network(network=net, device_name="CPU")
    return exec_net


def do_inference(exec_net, image):
    input_blob = next(iter(exec_net.inputs))
    return exec_net.infer({input_blob: image})


def get_crop_face(detections, image):
    w, h = image.shape[0], image.shape[1]
    ymin = int(detections[0, 0] * w)
    xmin = int(detections[0, 1] * h)
    ymax = int(detections[0, 2] * w)
    xmax = int(detections[0, 3] * h)
    margin_x = int(0.25 * (xmax - xmin))
    margin_y = int(0.25 * (ymax - ymin))
    ymin -= margin_y
    ymax += margin_y
    xmin -= margin_x
    xmax += margin_x
    face_img = image[ymin:ymax, xmin:xmax]
    face_img = cv2.cvtColor(face_img, cv2.COLOR_BGR2RGB)
    image = cv2.rectangle(image, (xmin, ymin), (xmax, ymax), (0, 0, 255), 2)
    return xmin, ymin, face_img
# postprocessing for mesh


def get_mesh_face(detections, face_img, image, xmin, ymin):
    xscale, yscale = 192 / face_img.shape[1], 192 / face_img.shape[0]
    for i in range(detections.shape[0]):
        x, y = int(detections[i, 0] / xscale), int(detections[i, 1] / yscale)
        image = cv2.circle(image, (xmin + x, ymin + y), 1, (255, 0, 0), 1)


# load BlazeFace model
blaze_net = load_to_IE("blazeface.onnx")
# load FaceMesh model
mesh_net = load_to_IE("facemesh.onnx")
# we need dynamically generated key for fetching output tensor
blaze_outputs = list(blaze_net.outputs.keys())
mesh_outputs = list(mesh_net.outputs.keys())
# to reuse postprocessing from BlazeFace
blazenet = BlazeFace()
blazenet.load_anchors("anchors.npy")
videoCapture = cv2.VideoCapture(0)
while True:
    ret, image = videoCapture.read()
    if not ret:
        break
    # get face detection boxes------------------------------------------------------------------
    # preprocessing
    face_img = cv2.dnn.blobFromImage(
        image, 1./127.5, (128, 128), (1, 1, 1), True)
    # inference
    output = do_inference(blaze_net, image=face_img)
    # postprocessing
    boxes = output[blaze_outputs[0]]
    confidences = output[blaze_outputs[1]]
    detections = blazenet._tensors_to_detections(
        boxes, confidences, blazenet.anchors)
    detections = np.squeeze(detections, axis=0)
    # take boxes
    xmin, ymin, face_img = get_crop_face(detections, image)
    # get face mesh ----------------------------------------------------------------------------
    # preprocessing
    mesh_img = cv2.dnn.blobFromImage(
        face_img, 1./127.5, (192, 192), (1, 1, 1), True)
    # inference
    output = do_inference(mesh_net, image=mesh_img)
    # postprocessing
    detections = output[mesh_outputs[1]].reshape(-1, 3)
    # take mesh
    get_mesh_face(detections, face_img, image, xmin, ymin)
    # show processed image
    cv2.imshow("capture", image)
    if cv2.waitKey(3) & 0xFF == 27:
        break



from typing import Union
from wanwu.core.enums import Backends
from wanwu.utils.image_utils import load_inputs_naive
from .det import Det
import numpy as np
from PIL import Image

try:
    import cv2

    has_cv2 = True
except ImportError:
    has_cv2 = False


class FaceWanwu(Det):

    def __init__(
        self,
        onnx_f: str = None,
        backend: Backends = Backends.CPU_ORT,
        type: str = "yolov5n-0.5",
        batch_size=1,
        input_width=416,
        input_height=416,
        num_classes=1,
        score_thr=0.4,
        nms_thr=0.45,
        keep_to_topk=100,
        fp16=False,
        timing=False,
        do_nms=True,
        **kargs
    ) -> None:
        super().__init__(
            onnx_f,
            backend,
            type,
            batch_size,
            input_width,
            input_height,
            num_classes,
            score_thr,
            nms_thr,
            keep_to_topk,
            fp16,
            timing,
            do_nms,
            **kargs
        )
        pass

    def get_face_boxes(self, im: Union[str, np.ndarray, Image.Image]) -> np.ndarray:

        if isinstance(im, str):
            if has_cv2:
                try:
                    im = cv2.imread(im)
                    im = cv2.cvtColor(im, cv2.COLOR_BGR2RGB)
                except Exception as e:
                    im = Image.open(im).convert("RGB")
            else:
                im = Image.open(im).convert("RGB")

            if isinstance(im, Image.Image):
                im = np.array(im)
        elif isinstance(im, Image.Image):
            im = np.array(im)
        inp = self.load_inputs(im, normalize_255=True)
        raw_boxes, scores, labels = self.infer(inp)
        return raw_boxes, scores

    def load_inputs(self, im, normalize_255=False):
        self.crt_image_raw_h = im.shape[0]
        self.crt_image_raw_w = im.shape[1]
        return load_inputs_naive(
            im,
            self.input_height,
            self.input_width,
            normalize_255=normalize_255,
        )



from enum import Enum


class Backends(Enum):
    GPU_TENSORRT = 1
    CPU_XNNPACK = 2
    CPU_OPENVINO = 3
    CPU_TVM = 4
    CPU_ORT = 5
    AUTO = 6
    GPU_ORT = 7
    CPU_MNN = 8
    CPU_PPLNN = 9
    GPU_PPLNN = 10

    @staticmethod
    def from_str(s: str):
        if s.lower() == 'tensorrt' or s.lower() == 'trt':
            return Backends.GPU_TENSORRT
        elif 'ort' in s.lower() or 'onnxruntime' in s.lower():
            if 'gpu' in s.lower():
                return Backends.GPU_ORT
            else:
                return Backends.CPU_ORT
        elif s.lower() == 'tvm':
            return Backends.CPU_TVM
        elif s.lower() == 'openvino':
            return Backends.CPU_OPENVINO
        elif s.lower() == 'auto':
            return Backends.AUTO
        elif s.lower() == 'gpu_ort':
            return Backends.GPU_ORT
        elif s.lower() == 'mnn':
            return Backends.CPU_MNN
        elif s.lower() == 'pplnn_cpu':
            return Backends.CPU_PPLNN
        elif s.lower() == 'pplnn_gpu':
            return Backends.GPU_PPLNN
        else:
            return Backends.AUTO

    def __str__(self) -> str:
        if self.value == 1:
            return 'trt'
        elif self.value == 2:
            return 'cpu'
        elif self.value == 5:
            return 'ort'
        elif self.value == 6:
            return 'auto'
        elif self.value == 7:
            return 'gpu_ort'
        elif self.value == 9:
            return 'ppl_cpu'
        elif self.value == 10:
            return 'ppl_gpu'
        else:
            return 'cpu'


# -*- coding: utf-8 -*-
# file: setup.py
# author: JinTian
# time: 12/02/2021 12:16 PM
# Copyright 2022 JinTian. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ------------------------------------------------------------------------

from setuptools import setup, find_packages

version_file = 'wanwu/version.py'


def get_version():
    with open(version_file, 'r') as f:
        exec(compile(f.read(), version_file, 'exec'))
    return locals()['__version__']


setup(name='wanwu',
      version=get_version(),
      keywords=['deep learning', 'script helper', 'tools'],
      description='Wanwu: Everything can have intelligence',
      long_description='''
      wanwu enables fastest inference on your target device by intelli switch TensorRT, OpenVINO, tvm etc automatically.
      also provides many model inference out-of-box without any further dependecies or development.
      ''',
      license='GPL-3.0',
      packages=[
          'wanwu.core',
          'wanwu.core.backends',
          'wanwu.det',
          'wanwu.seg',
          'wanwu.kps',
          'wanwu.cls',
          'wanwu.utils',
          'wanwu.track',
          'wanwu.base',
          'wanwu'
      ],
      entry_points={
          'console_scripts': [
              'wanwu = wanwu.wanwu:main'
          ]
      },
      author="Lucas Jin",
      author_email="jinfagang19@163.com",
      url='https://github.com/jinfagang/wanwu',
      platforms='any',
      install_requires=['colorama', 'requests', 'numpy', 'future', 'onnx',
                        'deprecated', 'alfred-py', 'onnxruntime', 'tabulate']
      )



from enum import Enum
from numpy import log
from wanwu.core.backends.openvino import OpenVINOWrapper

from wanwu.core.backends.ort import ORTWrapper
from wanwu.core.backends.mnn import MNNWrapper
from .graphsurgeon import OnnxModelSurgeonOperationer
from .infer import Backends
from alfred.utils.log import logger

try:
    from .backends.trt import TensorRTInferencer
    from alfred.deploy.tensorrt.common import build_engine_onnx_v2

    HAS_TRT = True
except ImportError as e:
    logger.warning(f'No TensorRT env found, disable TensorRT backend. {e}')
    HAS_TRT = False
import os


class TaskType(Enum):

    DET = 0
    SEG = 1
    KPS = 2
    CLS = 3
    BASE = 4

    @staticmethod
    def from_str(s: str):
        if s.lower() == "det":
            return TaskType.DET
        elif s.lower() == "seg":
            return TaskType.SEG
        elif s.lower() == "kps":
            return TaskType.KPS
        elif s.lower() == "cls":
            return TaskType.CLS
        elif s.lower() == "base":
            return TaskType.BASE
        else:
            return TaskType.CLS


class ModelLoader:
    """
    Loading model from onnx, then decide by backend how to construct
    an inference engine:

    if det model:
        if TensorRT:
            load model, inject batchedNMS plugin
        elif CPU:
            do normal load
    if seg model:
        normal
    """

    def __init__(
        self,
        onnx_f,
        backend: Backends,
        task_type: TaskType = TaskType.DET,
        fp16=False,
        int8=False,
        num_classes=80,
        batch_size=1,
        input_width=640,
        input_height=512,
        score_thr=0.2,
        nms_thr=0.6,
        keep_to_topk=100,
        nms_plugin_name="efficient_nms",
    ) -> None:
        self.onnx_f = onnx_f
        self.backend = backend      
        if os.path.basename(self.onnx_f).split(".")[-1] == "onnx":
            if task_type == TaskType.DET:
                if backend == Backends.GPU_TENSORRT:
                    self.save_p = self.get_trt_engine_save_f(onnx_f, fp16=fp16)

                    if not os.path.exists(self.save_p):
                        logger.info("Applying NMSPlugin to onnx for TensorRT deploy...")
                        self.onnx_model_gs = OnnxModelSurgeonOperationer(
                            self.onnx_f, num_classes=num_classes, batch_size=batch_size
                        )
                        self.onnx_model_gs.update_inputs_dims(
                            batch_size, input_height, input_width
                        )
                        if nms_plugin_name == "batched_nms":
                            logger.info(
                                "using BatchedNMSPlugin, this will be replaced in the future."
                            )
                            self.onnx_model_gs.register_trt_batched_nms(
                                score_thr, nms_thr, keep_to_topk
                            )
                        else:
                            self.onnx_model_gs.register_trt_efficient_nms(
                                score_thr, nms_thr, keep_to_topk
                            )
                        # BatchedNMS doesn't support dynamic HW input now, fix input in ONNX model
                        output_onnx_f, self.onnx_model = self.onnx_model_gs.save()
                        logger.info(
                            "An onnx model with nms op saved into: {}".format(
                                output_onnx_f
                            )
                        )

                        if self.onnx_model_gs.is_dynamic_on_hw:
                            opt_params = {
                                "images": [
                                    [
                                        batch_size,
                                        3,
                                        input_height,
                                        input_width,
                                    ],  # min shape
                                    [
                                        batch_size,
                                        3,
                                        input_height,
                                        input_width,
                                    ],  # opt shape
                                    [
                                        batch_size,
                                        3,
                                        input_height,
                                        input_width,
                                    ],  # max shape
                                ]
                            }
                        else:
                            opt_params = None

                        logger.info(
                            "Start build TensorRT engine, this will only run once."
                        )
                        e = build_engine_onnx_v2(
                            onnx_file_path=output_onnx_f,
                            engine_file_path=self.save_p,
                            fp16_mode=fp16,
                            save_engine=True,
                            opt_params=opt_params,
                        )
                        if e is not None:
                            logger.info("engine saved into: {}".format(self.save_p))
                        else:
                            RuntimeError("Failed to build engine, aborted.")

                else:
                    logger.info(
                        f"Determine ONNX input with specified: {input_height}x{input_width}"
                    )
                    self.onnx_model_gs = OnnxModelSurgeonOperationer(
                        self.onnx_f, num_classes=num_classes
                    )
                    if (
                        not self.onnx_model_gs.is_dynamic_on_hw
                        and self.onnx_model_gs.ok
                    ):
                        self.onnx_model_gs.update_inputs_dims(
                            batch_size, input_height, input_width
                        )
                        self.onnx_model_gs.save(self.onnx_f)
                        logger.info("New input dimension ONNX saved.")
                    # normally load onnx model
                    self.onnx_model = None
            elif task_type == TaskType.CLS or task_type == TaskType.KPS:
                if backend == Backends.GPU_TENSORRT:
                    self.save_p = self.get_trt_engine_save_f(onnx_f, fp16=fp16)

                    if not os.path.exists(self.save_p):
                        logger.info("Surgeon on ONNX model for TensorRT deploy...")
                        self.onnx_model_gs = OnnxModelSurgeonOperationer(
                            self.onnx_f, num_classes=num_classes
                        )
                        self.onnx_model_gs.update_inputs_dims(
                            batch_size, input_height, input_width
                        )
                        output_onnx_f, self.onnx_model = self.onnx_model_gs.save(
                            output_path=self.onnx_f
                        )
                        logger.info(
                            "An onnx model saved into: {}".format(output_onnx_f)
                        )
                        if self.onnx_model_gs.is_dynamic_on_hw:
                            opt_params = {
                                "images": [
                                    [
                                        batch_size,
                                        3,
                                        input_height,
                                        input_width,
                                    ],  # min shape
                                    [
                                        batch_size,
                                        3,
                                        input_height,
                                        input_width,
                                    ],  # opt shape
                                    [
                                        batch_size,
                                        3,
                                        input_height,
                                        input_width,
                                    ],  # max shape
                                ]
                            }
                        elif self.onnx_model_gs.is_dynamic_on_batch:
                            input_specs = self.onnx_model_gs.inputs_dict
                            logger.info(
                                f"model is dynamic on batch, try adding opt params for trt: {input_specs}"
                            )
                            ss = list(input_specs.values())[0][1:]
                            # TODO: this is hard coded
                            opt_params = {
                                list(input_specs.keys())[0]: [
                                    [1, ss[0], ss[1], ss[2]],  # min shape
                                    [16, ss[0], ss[1], ss[2]],  # opt shape
                                    [32, ss[0], ss[1], ss[2]],  # max shape
                                ]
                            }
                        else:
                            opt_params = None
                            logger.info(
                                "onnx model is not dynamic, directly convert to trt engine."
                            )

                        logger.info(
                            "Start build TensorRT engine, this will only run once."
                        )
                        e = build_engine_onnx_v2(
                            onnx_file_path=self.onnx_f,
                            engine_file_path=self.save_p,
                            fp16_mode=fp16,
                            save_engine=True,
                            opt_params=opt_params,
                        )
                        if e is not None:
                            logger.info("engine saved into: {}".format(self.save_p))
                        else:
                            RuntimeError("Failed to build engine, aborted.")
                else:
                    logger.info(
                        f"Determine ONNX input with specified: {input_height}x{input_width}"
                    )
                    self.onnx_model_gs = OnnxModelSurgeonOperationer(
                        self.onnx_f, num_classes=num_classes
                    )
                    if (
                        not self.onnx_model_gs.is_dynamic_on_hw
                        and self.onnx_model_gs.ok
                    ):
                        self.onnx_model_gs.update_inputs_dims(
                            batch_size, input_height, input_width
                        )
                        self.onnx_model_gs.save(self.onnx_f)
                        logger.info("New input dimension ONNX saved.")
                    # normally load onnx model
                    self.onnx_model = None
            elif task_type == TaskType.SEG:
                self.onnx_model_gs = OnnxModelSurgeonOperationer(
                    self.onnx_f, num_classes=num_classes
                )
                self.is_dynamic_on_hw = self.onnx_model_gs.is_dynamic_on_hw
                if (
                    self.onnx_model_gs.is_dynamic_on_hw
                    and input_width != None
                    and input_height != None
                ):
                    self.onnx_model_gs.update_inputs_dims(
                        batch_size, input_height, input_width
                    )
                    self.onnx_model_gs.save(self.onnx_f)
                    logger.info("New input dimension ONNX saved.")
                # normally load onnx model
                self.onnx_model = None
            elif task_type == TaskType.BASE:
                logger.info("TaskBase will pass onnx graph surgeon.")
                if backend == Backends.GPU_TENSORRT:
                    self.save_p = self.get_trt_engine_save_f(onnx_f, fp16=fp16)
                    logger.info(f'will save engine to: {self.save_p}')
                    if not os.path.exists(self.save_p):
                        logger.info("Surgeon on ONNX model for TensorRT deploy...")
                        self.onnx_model_gs = OnnxModelSurgeonOperationer(
                            self.onnx_f, num_classes=num_classes
                        )
                        self.onnx_model_gs.update_inputs_dims(
                            batch_size, input_height, input_width
                        )
                        output_onnx_f, self.onnx_model = self.onnx_model_gs.save(
                            output_path=self.onnx_f
                        )
                        logger.info(
                            "An onnx model saved into: {}".format(output_onnx_f)
                        )
                        if self.onnx_model_gs.is_dynamic_on_hw:
                            opt_params = {
                                "images": [
                                    [
                                        batch_size,
                                        3,
                                        input_height,
                                        input_width,
                                    ],  # min shape
                                    [
                                        batch_size,
                                        3,
                                        input_height,
                                        input_width,
                                    ],  # opt shape
                                    [
                                        batch_size,
                                        3,
                                        input_height,
                                        input_width,
                                    ],  # max shape
                                ]
                            }
                        elif self.onnx_model_gs.is_dynamic_on_batch:
                            input_specs = self.onnx_model_gs.inputs_dict
                            logger.info(
                                f"model is dynamic on batch, try adding opt params for trt: {input_specs}"
                            )
                            ss = list(input_specs.values())[0][1:]
                            # TODO: this is hard coded
                            opt_params = {
                                list(input_specs.keys())[0]: [
                                    [1, ss[0], ss[1], ss[2]],  # min shape
                                    [16, ss[0], ss[1], ss[2]],  # opt shape
                                    [32, ss[0], ss[1], ss[2]],  # max shape
                                ]
                            }
                        else:
                            opt_params = None
                            logger.info(
                                "onnx model is not dynamic, directly convert to trt engine."
                            )
                        logger.info(
                            "Start build TensorRT engine, this will only run once."
                        )
                        e = build_engine_onnx_v2(
                            onnx_file_path=self.onnx_f,
                            engine_file_path=self.save_p,
                            fp16_mode=fp16,
                            save_engine=True,
                            opt_params=opt_params,
                        )
                        if e is not None:
                            logger.info("engine saved into: {}".format(self.save_p))
                        else:
                            RuntimeError("Failed to build engine, aborted.")
            else:
                logger.info(f"unsupported task now: {task_type}")
        else:
            logger.info("model not in onnx format, pass preprocess onnx step.")

    def get_trt_engine_save_f(self, onnx_f, fp16=False, int8=False):
        if fp16:
            save_p = os.path.join(
                os.path.dirname(onnx_f),
                os.path.basename(onnx_f).replace(".onnx", "_fp16.engine"),
            )
        elif int8:
            save_p = os.path.join(
                os.path.dirname(onnx_f),
                os.path.basename(onnx_f).replace(".onnx", "_int8.engine"),
            )
        else:
            save_p = os.path.join(
                os.path.dirname(onnx_f),
                os.path.basename(onnx_f).replace(".onnx", ".engine"),
            )
        return save_p

    def get_infer_wrapper(self):
        if (
            os.path.basename(self.onnx_f).split(".")[-1] == "mnn"
            and self.backend != Backends.CPU_MNN
        ):
            logger.info(
                f"detected mnn file, but backend is: {self.backend.name}, force to MNN backend."
            )
            self.backend = Backends.CPU_MNN
            return MNNWrapper(self.onnx_f)
        else:
            if self.backend == Backends.GPU_TENSORRT and HAS_TRT:
                return TensorRTInferencer(self.save_p)
            elif self.backend == Backends.CPU_ORT:
                # we don't do graphsurgeon for now.
                return ORTWrapper(self.onnx_f)
            elif (
                self.backend == Backends.CPU_MNN
                and os.path.basename(self.onnx_f).split(".")[-1] == "mnn"
            ):
                # we don't do graphsurgeon for now.
                return MNNWrapper(self.onnx_f)
            elif self.backend == Backends.CPU_OPENVINO:
                return OpenVINOWrapper(self.onnx_f)
            else:
                return ORTWrapper(self.onnx_f)



from abc import ABC, abstractclassmethod
from turtle import back
from .cellar import model_urls, LOCAL_CACHE_FOLDER
import onnx
from alfred.utils.file_io import download
from .dispatcher import dispatcher_on_device_auto
from .enums import Backends
from alfred.utils.log import logger
import os


AVAILABLE_BACKENDS = [
    Backends.GPU_TENSORRT,
    Backends.CPU_OPENVINO,
    Backends.CPU_ORT,
    Backends.AUTO,
    Backends.CPU_MNN,
    Backends.CPU_PPLNN,
]


def get_model_name_suffix_by_backend(backend: Backends):
    if backend in [Backends.GPU_ORT]:
        return "gpu"
    elif backend == Backends.GPU_TENSORRT:
        return "trt"
    elif backend in [Backends.CPU_ORT, Backends.CPU_OPENVINO]:
        return "cpu"
    else:
        return 'cpu'


def get_model_path_by_key(model_type, backend):
    model_type = model_type + f"_{get_model_name_suffix_by_backend(backend)}"
    assert model_type in model_urls.keys(), f"{model_type} not in warehouse"
    if model_type in model_urls.keys():
        f_url = model_urls[model_type]
        f_path = os.path.join(LOCAL_CACHE_FOLDER, f_url.split("/")[-1])
        if not os.path.exists(f_path):
            logger.info(f"downloading model from: {f_url} into {LOCAL_CACHE_FOLDER}")
            f_path = download(f_url, LOCAL_CACHE_FOLDER)
        return f_path
    else:
        return None


class BaseInfer(ABC):
    def __init__(
        self,
        type: str,
        onnx_f: str,
        backend: Backends,
        batch_size=1,
        input_width=640,
        input_height=512,
    ) -> None:
        super().__init__()
        self.type = type

        self.backend = backend
        assert (
            self.backend in AVAILABLE_BACKENDS
        ), f"{self.backend} not in currently supported backends: {AVAILABLE_BACKENDS}"

        if self.backend == Backends.AUTO:
            logger.info("Backend in AUTO mode, deciding on auto device.")
            self.backend = dispatcher_on_device_auto()

        if onnx_f is not None:
            self.onnx_f = onnx_f
        else:
            if type is not None:
                # get models from type
                type = type + f"_{get_model_name_suffix_by_backend(self.backend)}"
                assert (
                    type in model_urls.keys()
                ), f"unsupported model built in: {type}, avaiable: {model_urls.keys()}"
                f_url = model_urls[type]
                f_path = os.path.join(LOCAL_CACHE_FOLDER, f_url.split("/")[-1])
                if not os.path.exists(f_path):
                    logger.info(
                        f"downloading model from: {f_url} into {LOCAL_CACHE_FOLDER}"
                    )
                    f_path = download(f_url, LOCAL_CACHE_FOLDER)
                self.onnx_f = f_path
            else:
                ValueError(
                    "Either onnx_f or type not specific, I dont know what model you want run."
                )

        assert (
            self.onnx_f is not None
        ), "Can not either get onnx file from type or onnx_f, please set at least one of them."

        self.batch_size = batch_size
        self.input_width = input_width
        self.input_height = input_height

    @abstractclassmethod
    def infer(self, inp):
        NotImplemented

    @abstractclassmethod
    def vis_res(self, res):
        NotImplemented



from .base import TensorInfo, WrapperBase
import numpy as np
from alfred.utils.log import logger
import os
import sys

try:
    from pyppl import nn as pplnn
    from pyppl import common as pplcommon
except ImportError:
    pplnn = None


class PPLModel(object):
    def __init__(self, model_file, engine_type="x86"):
        self._engines = []
        if engine_type == "x86":
            self._create_x86_engine()
        elif engine_type == "cuda":
            self._create_cuda_engine()
        else:
            logger.error("not support engine type: ", engine_type)
            sys.exit(-1)
        self._create_runtime(model_file)

    def _create_x86_engine(self):
        x86_options = pplnn.x86.EngineOptions()
        x86_options.mm_policy = pplnn.x86.MM_COMPACT
        x86_engine = pplnn.x86.EngineFactory.Create(x86_options)
        if not x86_engine:
            logger.error("create x86 engine failed.")
            sys.exit(-1)
        self._engines.append(x86_engine)

    def _create_cuda_engine(self):
        cuda_options = pplnn.cuda.EngineOptions()
        cuda_options.device_id = 0
        cuda_engine = pplnn.cuda.EngineFactory.Create(cuda_options)
        self._engines.append(cuda_engine)

    def _create_runtime(self, model_file_name):
        runtime_builder = pplnn.onnx.RuntimeBuilderFactory.Create()
        if not runtime_builder:
            logger.error("create RuntimeBuilder failed.")
            sys.exit(-1)

        status = runtime_builder.LoadModelFromFile(model_file_name)
        if status != pplcommon.RC_SUCCESS:
            logger.error(
                "init OnnxRuntimeBuilder failed: " + pplcommon.GetRetCodeStr(status)
            )
            sys.exit(-1)

        resources = pplnn.onnx.RuntimeBuilderResources()
        resources.engines = self._engines
        status = runtime_builder.SetResources(resources)
        if status != pplcommon.RC_SUCCESS:
            logger.error(
                "OnnxRuntimeBuilder SetResources failed: "
                + pplcommon.GetRetCodeStr(status)
            )
            sys.exit(-1)

        status = runtime_builder.Preprocess()
        if status != pplcommon.RC_SUCCESS:
            logger.error(
                "OnnxRuntimeBuilder preprocess failed: "
                + pplcommon.GetRetCodeStr(status)
            )
            sys.exit(-1)

        self._runtime = runtime_builder.CreateRuntime()
        if not self._runtime:
            logger.error("create Runtime instance failed.")
            sys.exit(-1)

    def _prepare_input(self, input_file):
        """set input data"""
        tensor = self._runtime.GetInputTensor(0)
        in_data = np.fromfile(input_file, dtype=np.float32).reshape((1, 3, 800, 1200))
        status = tensor.ConvertFromHost(in_data)
        if status != pplcommon.RC_SUCCESS:
            logger.error(
                "copy data to tensor["
                + tensor.GetName()
                + "] failed: "
                + pplcommon.GetRetCodeStr(status)
            )
            sys.exit(-1)

    def prepare_input_from_numpy(self, inputs):
        """set input data"""
        if isinstance(inputs, dict):
            for i in range(self._runtime.GetInputCount()):
                tensor = self._runtime.GetInputTensor(i)
                assert (
                    tensor.GetName() in inputs.keys()
                ), f"can not find input: {tensor.GetName()}"
                in_data = np.array(inputs[tensor.GetName()], dtype=np.float32)
                status = tensor.ConvertFromHost(in_data)
                if status != pplcommon.RC_SUCCESS:
                    logger.error(
                        "copy data to tensor["
                        + tensor.GetName()
                        + "] failed: "
                        + pplcommon.GetRetCodeStr(status)
                    )
                    sys.exit(-1)
        else:
            tensor = self._runtime.GetInputTensor(0)
            in_data = np.array(inputs, dtype=np.float32)
            status = tensor.ConvertFromHost(in_data)
            if status != pplcommon.RC_SUCCESS:
                logger.error(
                    "copy data to tensor["
                    + tensor.GetName()
                    + "] failed: "
                    + pplcommon.GetRetCodeStr(status)
                )
                sys.exit(-1)

    def _prepare_output(self):
        """save output"""
        res = {}
        for i in range(self._runtime.GetOutputCount()):
            tensor = self._runtime.GetOutputTensor(i)
            tensor_data = tensor.ConvertToHost()
            if not tensor_data:
                logger.error("copy data from tensor[" + tensor.GetName() + "] failed.")
                sys.exit(-1)
            res[tensor.GetName()] = np.array(tensor_data, copy=False)
        return res

    def run(self):
        """run pplmodel

        Keyword arguments:
        engine_type -- which engine to use x86 or cuda
        model_file_name -- input model file
        input_file -- input data file (binary data)
        """

        status = self._runtime.Run()
        if status != pplcommon.RC_SUCCESS:
            logger.error("Run() failed: " + pplcommon.GetRetCodeStr(status))
            sys.exit(-1)
        return self._prepare_output()

    def get_inputs_dict(self):
        res = {}
        for i in range(self._runtime.GetInputCount()):
            tensor = self._runtime.GetInputTensor(i)
            res[tensor.GetName()] = TensorInfo(tensor.GetShape(), tensor.GetDataType())
        return res

    def get_outputs_dict(self):
        res = {}
        for i in range(self._runtime.GetOutputCount()):
            tensor = self._runtime.GetOutputTensor(i)
            res[tensor.GetName()] = TensorInfo(tensor.GetShape(), tensor.GetDataType())
        return res


class PPLCPUWrapper(WrapperBase):
    def __init__(self, onnx_f) -> None:
        """
        PPL.NN inference as backend
        """
        if pplnn is None:
            logger.error("import pyppl failed, ppl backend not available.")
        else:
            self.onnx_f = onnx_f
            logger.info(f"Loading from onnx model: {self.onnx_f}")

            # online parse onnx file to mnn format?
            self.ppl_model = PPLModel(self.onnx_f, "x86")

    def infer_si(self, inputs):
        if isinstance(inputs, list):
            inputs = np.array(inputs)

        self.ppl_model.prepare_input_from_numpy(inputs)
        res = self.ppl_model.run()
        return res

    def infer(self, inputs, return_dict=False):
        if isinstance(inputs, list):
            inputs = np.array(inputs)

        self.ppl_model.prepare_input_from_numpy(inputs)
        res = self.ppl_model.run()
        return res

    def get_inputs_dict(self):
        return self.ppl_model.get_inputs_dict()

    def get_outputs_dict(self):
        return self.ppl_model.get_outputs_dict()
